import math
from copy import deepcopy
from ffxivcalc.helperCode.Vocal import PrintResult, SimulateRuns
from ffxivcalc.Jobs.PlayerEnum import *
from ffxivcalc.Jobs.ActionEnum import name_for_id
from ffxivcalc.Jobs.Base_Spell import ZIPAction, PreBakedAction
from ffxivcalc.helperCode.Progress import ProgressBar
from ffxivcalc.SimulationRecord.record import SimulationRecord, page
from ffxivcalc.helperCode.exceptions import playerIDNotFound

from ffxivcalc.Jobs.Base_Spell import WaitAbility

import matplotlib.pyplot as plt
import logging
from ffxivcalc.helperCode.helper_math import roundDown, isclose, roundUp
main_logging = logging.getLogger("ffxivcalc")
fight_logging = main_logging.getChild("Fight")


class NoMoreAction(Exception):# Exception called if a spell fails to cast
    pass

class Fight:
    """
    
    This class will be the environment in which the fight happens. It will hold a list of players, an enemy, etc.
    It will be called upon for when we want to start the simulation

    """

    def getPlayerPrepullLength(self, ignoreFirstStandardFinish : bool = False, ignoreFirstTechnicalFinish : bool = False) -> dict[int : float]:
        """
        Computes and returns the length of the prepull of all players. Returns a dictionnary that maps playerID to prepull length.

        ignoreFirstStandardFinish : bool - If true assumes first standard finish was done out of range (so didn't start fight).
        ignoreFirstTechnicalFinish : bool - If true assumes first technical finish was done out of range (so didn't start fight).

        """

        prepullLengthDict = {}

        for player in self.PlayerList:
            prepullLengthDict[player.playerID] = player.getPlayerPrePullTime(ignoreFirstStandardFinish=ignoreFirstStandardFinish, ignoreFirstTechnicalFinish=ignoreFirstTechnicalFinish)

        return prepullLengthDict
    
    def syncPlayerPrePull(self, editActionSet : bool = True) -> dict[int : float]:
        """
        Computes the time difference between all player's prepull and makes sure they all match and all players start doing damage at the same time.
        
        Returns a dictionnary that maps playerID to the added waitTime to sync all players.

        editActionSet : bool - If set to true will insert at index 0 a waitAbility of the needed time.
        """


        prepullLengthDict = self.getPlayerPrepullLength()

        fight_logging.debug(f'Synchronising player prepull length. \nCurrent prepull length : {prepullLengthDict}')

        maxPrepullPlayerID = max(prepullLengthDict.keys(), key=lambda x : prepullLengthDict[x])

        fight_logging.debug(f'Detected max length : {maxPrepullPlayerID}')

        addedAmountDict = {}

        for player in self.PlayerList:
            if maxPrepullPlayerID == player.playerID : continue

            amountToAdd = prepullLengthDict[maxPrepullPlayerID] - prepullLengthDict[player.playerID]
            addedAmountDict[player.playerID] = round(amountToAdd,2)
            fight_logging.debug(f'Adding waitAbility({amountToAdd}) to player {player.playerID}')

            if editActionSet: player.ActionSet.insert(0, WaitAbility(amountToAdd))

        verifDict = self.getPlayerPrepullLength()
        fight_logging.debug(f'Synchronising done. {verifDict}')

        return addedAmountDict

    def GetEnemityList(self, range : int):
        """Returns a list of players in order of greather enemity to lowest enemity. The length of the
        returned list is equal to range.

        range (int) : Number of players we want
        
        """


        sorted_list = sorted(self.PlayerList, key=lambda Player : Player.TotalEnemity, reverse=True)
        # This returns a sorted list of the PlayerList with respect to their TotalEnemity
        # This could be optimized. But for now we will reocompute this sorted list every time.

        return sorted_list[0:range] # Returns the number of targets we are interested in

    def playerForID(self, id):
        """This function returns a reference to the player object that has the given ID in the fight.

        Args:
            id (int): id of the player we want.
        """

        for player in self.PlayerList:
            if player.playerID == id : return player

        raise playerIDNotFound
    
    def deepCopy(self):
        """This function returns a deepcopy of itself.
        We cannot use the deepcopy function alone since some issues arise.
        One such issue is that any action that targets a player will not have their target
        updated and will still target the player from the original fight (the one that is being deepcopied).
        To fix this we will look for the actions with target, get the targetID and replace these actions with the appropriate
        action that now targets the correct player.
        Note that because of how this is implemented, any fight object that has a buff given to another player must have both the caster
        and the target in its player list as if the target is not in the playerList it will not receive a valid ID.
        """
        from ffxivcalc.Jobs.Ranged.Dancer.Dancer_Spell import ClosedPosition
        from ffxivcalc.Jobs.Melee.Dragoon.Dragoon_Spell import DragonSight
        from ffxivcalc.Jobs.Healer.Astrologian.Astrologian_Spell import Balance, Arrow, Spear, Bole, Ewer, Spire

        dcFight = deepcopy(self)

        for player in dcFight.PlayerList:
            match player.JobEnum:
                case JobEnum.Dancer:
                    changeIndex = []
                    
                    for index,action in enumerate(player.ActionSet):
                        if action.id == 16006: # Closed Position
                            changeIndex.append((index,action.TargetPlayerObject.playerID))
                    for change in changeIndex:
                        player.ActionSet[change[0]] = ClosedPosition(dcFight.playerForID(change[1]))

                case JobEnum.Dragoon:
                    changeIndex = []
                    
                    for index,action in enumerate(player.ActionSet):
                        if action.id == 7398: # Dragoon Sight
                            changeIndex.append((index,action.TargetPlayerObject.playerID))
                    for change in changeIndex:
                        player.ActionSet[change[0]] = DragonSight(dcFight.playerForID(change[1]))
                case JobEnum.Astrologian:
                    changeIndex = []
                    
                    for index,action in enumerate(player.ActionSet):
                        if action.id >=4401 and action.id <=4406: # Arcanum
                            match action.id :
                                case 4401 : changeIndex.append((index,Balance(dcFight.playerForID(action.TargetPlayerObject.playerID))))
                                case 4402 : changeIndex.append((index,Arrow(dcFight.playerForID(action.TargetPlayerObject.playerID))))
                                case 4403 : changeIndex.append((index,Spear(dcFight.playerForID(action.TargetPlayerObject.playerID))))
                                case 4404 : changeIndex.append((index,Bole(dcFight.playerForID(action.TargetPlayerObject.playerID))))
                                case 4405 : changeIndex.append((index,Ewer(dcFight.playerForID(action.TargetPlayerObject.playerID))))
                                case 4406 : changeIndex.append((index,Spire(dcFight.playerForID(action.TargetPlayerObject.playerID))))

                    for change in changeIndex:
                        player.ActionSet[change[0]] = change[1]

        return dcFight

    def __init__(self, Enemy, ShowGraph):
        self.Enemy = Enemy
        Enemy.CurrentFight = self
        self.ShowGraph = ShowGraph
        self.TimeStamp = 0
        self.TeamCompositionBonus = 1
        self.FirstHit = False # False until the first damaging action is done
        self.RequirementOn = True # By default True
        self.FightStart = False # If Fight is started
        self.IgnoreMana = True # If true we ignore mana in the simulation
        self.timeValue = [] # Array holding all sampling time for the DPS and PPS
        self.failedRequirementList = [] # List holding all failedRequirementEvent object for the fight.
        self.waitingThreshold = 1 # number of seconds we are willing to wait for. By default 1
        self.wipe = False # Will be set to True in case we are stopping the simulation.
        self.PlayerList = [] # Empty player list
        self.MaxPotencyPlentifulHarvest = False # True will make Plentiful Harvest do max potency regardless of player.
        self.TimeUnit = 0
        self.nextID = 1 # Used to give ID to players. ID 0 is invalid

        self.simulationRecord = SimulationRecord() # Creating SimulationRecord object

                             # These values can only be eddited by manually changing the values by accessing the Fight object.
        self.SavePreBakedAction = False
        self.PlayerIDSavePreBakedAction = 0
        self.alwaysAllowConditionalAction = False # Having this value set to True will make the conditionalActions always happen for players in the fight.

        # functions

        def DefaultNextActionFunction(Fight, Player) -> bool:
            """
            The default function does not add any other actions and hence just returns True to lock the player.
            """
            return True

        def DefaultExtractInfo(Fight) -> None:
            """
            This function is called every frame of the simulation and can be used to extract whatever information from it we want.
            By default it does nothing.
            Note that the function cannot return any value, and so it must save all wanted data in some other variable or file.
            """
            pass

        self.ComputeDamageFunction = ComputeDamage 
        self.ComputeHealingFunction = ComputeHeal
        self.NextActionFunction = DefaultNextActionFunction
        self.ExtractInfo = DefaultExtractInfo

    def AddPlayer(self, Players : list, giveNewID : bool = True):
        """This function adds the given player list to the fight's playerList.
        giveNewID : bool - If true will give a new ID
        """
        for player in Players:
            player.CurrentFight = self
            self.PlayerList.append(player)
            if giveNewID:
                player.playerID = self.nextID
                self.nextID += 1

    def SimulateZIPFight(self):
        """
        This function simulates the fight using the ZIPActionList of players. This will simulate a random DPS occurence
        and be used to compute a distribution.
        """

        for player in self.PlayerList:
            player_current_damage = 0
            for ZIPAction in player.ZIPActionSet:
                player_current_damage += ZIPAction.ComputeRandomDamage()
            player.ZIPDPSRun.append(round(player_current_damage/self.TimeStamp/20)*20)

    def SimulatePreBakedFight(self, Index : int, MainStat : int, f_WD : float, f_DET : float, f_TEN : float, f_SPD : float, f_CritRate : float, f_CritMult : float, f_DH : float, DHAuto : float, n : int = 1, getInfo : bool = False):
        """
        This function is called when the user wants to simulate the damage done by the pre baked actions. The player ID with
        the pre baked actions must be given. The user must also specify all the damage values computed from the stats.
        Index : int -> Index of the player with the PreBakedActions.
        n : int -> number of trial for random DPS simulation
        """
        player = self.PlayerList[Index]
        ExpectedDamage = 0
        baseMain = 390

        damageHistory = []   # This list will contain the damage of all PreBakedAction ComputeExpectedDamage used to
                             # faster compute RandomDamage

        countGCD = 0
        timeStamp = 0
        totalPotency = 0
        countGCD = 0
        #for history in player.PercentBuffHistory:
        #    fight_logging.warning(str(history))

                             # Will compute DPS
        for PreBakedAction in player.PreBakedActionSet:
                                                     # Count every GCD
            if PreBakedAction.isGCD : countGCD += 1                        
            timeStamp = PreBakedAction.timeStamp
                                         # Computing base MainStat for this action. If from pet do not get teamcomp bonus
            totalPotency += PreBakedAction.Potency
            curMainStat = MainStat * (PreBakedAction.MainStatPercentageBonus if not PreBakedAction.isFromPet else 1)

            fight_logging.debug("TimeStamp : " + str(timeStamp))
                                         # Potion
            if PreBakedAction.potionIsActive : curMainStat = math.floor(min(math.floor(curMainStat * 1.1), curMainStat + 262)) #Grade 8 HQ tincture
                                         # Any other percent bonus. Might have to remake dragoon tether since pet not affected by this buff

            for buff in PreBakedAction.buffList:
                PreBakedAction.PercentageBonus.append(buff.MultDPS)

            if PreBakedAction.IsTank : f_MAIN_DMG = (100+math.floor((curMainStat-baseMain)*156/baseMain))/100 # Tanks have a difference constant 
            else: f_MAIN_DMG = (100+math.floor((curMainStat-baseMain)*195/baseMain))/100
            ActionExpected, Damage = PreBakedAction.ComputeExpectedDamage(f_MAIN_DMG,f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto) #if not actionFailsCondition else (0,0)
            ExpectedDamage += ActionExpected
            player.DamageInstanceList.append(ActionExpected)
            damageHistory.append(Damage)

                             # Will compute Random DPS
        #randomDPSRuns = []   # This list will contain all the DPS of the random runs
        #for run in range(n):
        #    CurrentDamage = 0
        #    index = 0
        #    for PreBakedAction in player.PreBakedActionSet:
        #        CurrentDamage += PreBakedAction.ComputeRandomDamage(damageHistory[index], f_CritRate,f_CritMult, f_DH)
        #        index += 1 
        #        
        #        if index == len(damageHistory) : break # Might have to break if we ommit some autos.

        #    randomDPSRuns.append(CurrentDamage/timeStamp)
                             # Sorting array so we can find the percentiles.
        #randomDPSRuns.sort()

        #Percent = int(n/100)
        #percentileRuns = {
        #    "1" :  0 if n < 100 else sum(randomDPSRuns[(Percent):(10*Percent-1)])/(len(randomDPSRuns[(Percent+1):(10*Percent-1)])),
        #    "10" : 0 if n < 100 else sum(randomDPSRuns[(10*Percent+1):(25*Percent-1)])/(len(randomDPSRuns[(10*Percent+1):(25*Percent-1)])),
        #    "25" : 0 if n < 100 else sum(randomDPSRuns[(25*Percent+1):(50*Percent-1)])/(len(randomDPSRuns[(25*Percent+1):(50*Percent-1)])),
        #    "50" : 0 if n < 100 else sum(randomDPSRuns[(50*Percent+1):(75*Percent-1)])/(len(randomDPSRuns[(50*Percent+1):(75*Percent-1)])),
        #    "75" : 0 if n < 100 else sum(randomDPSRuns[(75*Percent+1):(90*Percent-1)])/(len(randomDPSRuns[(75*Percent+1):(90*Percent-1)])),
        #    "90" : 0 if n < 100 else sum(randomDPSRuns[(90*Percent+1):(99*Percent-1)])/(len(randomDPSRuns[(90*Percent+1):(99*Percent-1)])),
        #    "99" : 0 if n < 100 else sum(randomDPSRuns[(99*Percent):])/(len(randomDPSRuns[(99*Percent):]))
        #}

        for preBakedAction in player.PreBakedActionSet: preBakedAction.resetPercentageBonus()


        fight_logging.debug("counted GCD  : " + str(countGCD))
        fight_logging.debug("previous GCD : " + str(player.GCDCounter))
        if getInfo : return round(ExpectedDamage/timeStamp,2), ExpectedDamage, timeStamp, totalPotency
        return round(ExpectedDamage/timeStamp if True else ExpectedDamage,2), {}

    def SimulateFight(self, TimeUnit, TimeLimit, vocal, PPSGraph : bool = True, MaxTeamBonus : bool = False, MaxPotencyPlentifulHarvest : bool = False, n = 0, showProgress : bool = True,
                      computeGraph : bool = True, loadingBarBuffer = None) -> None:

        """
        This function will Simulate the fight given the enemy and player list of this Fight
        It will increment in TimeUnit up to a maximum of TimeLimit (there can be other reasons the Fight ends)
        It will check weither a player can cast its NextSpell, and if it can it will call the relevant functions
        However, no direct computation is done in this function, it simply orchestrates the whole thing

        TimeUnit : float -> unit at which the simulator will advance through time in the simulation
        TimeLimit : float -> time limit at which the simulator will stop
        vocal : bool -> True if we want to print out the results
        PPSGraph (bool) = True -> If we want the PPS graph to be next to the DPS graph
        MaaxTeamBonus (bool) = False -> If true, gives the 5% bonus regardless of team comp
        showProgress (bool) = True -> If true show fight progress bar. Still computes the Pb even if false.
        computeGraph (bool) = True -> If true will process the Graphs even if they do not show.
        loadingBarBuffer = None -> This dict will have the key 'pb' be set as the progress bar of the simulation
        
        """
        self.MaxPotencyPlentifulHarvest = MaxPotencyPlentifulHarvest
        self.TimeStamp = 0   # Keep track of the time
        start = False
        self.TimeUnit = TimeUnit

        self.showProgress = showProgress
        self.computeGraph = computeGraph


        self.timeValue = []  # Used for graph

        self.ComputeFunctions() # Compute all damage functions for the players


                             # Checks if every player has a least one action. If not
                             # we put True Lock
        for player in self.PlayerList:
            if len(player.ActionSet) == 0 : player.TrueLock = True

        # The first thing we will do is compute the TEAM composition DPS bonus
        # each class will give 1%
        # Tank, Healer, Caster, Ranged, Melee
        if MaxTeamBonus:
            self.TeamCompositionBonus = 1.05
        else:
            hasMelee = False
            hasCaster = False
            hasRanged = False
            hasTank = False
            hasHealer = False
            for player in self.PlayerList:
                if player.RoleEnum == RoleEnum.Melee : hasMelee = True
                elif player.RoleEnum == RoleEnum.PhysicalRanged : hasCaster = True
                elif player.RoleEnum == RoleEnum.Caster : hasRanged = True
                elif player.RoleEnum == RoleEnum.Healer : hasTank = True
                elif player.RoleEnum == RoleEnum.Tank : hasHealer = True

            if len(self.PlayerList) == 1 : self.TeamCompositionBonus = 1 # If only one player, there is not bonus
            else:
                if hasMelee: self.TeamCompositionBonus += 0.01
                if hasCaster: self.TeamCompositionBonus += 0.01
                if hasRanged: self.TeamCompositionBonus += 0.01
                if hasTank: self.TeamCompositionBonus += 0.01
                if hasHealer: self.TeamCompositionBonus += 0.01

        for Player in self.PlayerList:
            Player.Stat["MainStat"] = math.floor(Player.Stat["MainStat"] * self.TeamCompositionBonus) # Scaling %bonus on mainstat

        # Will first compute each player's GCD reduction value based on their Spell Speed and Skill Speed Value
        for Player in self.PlayerList:
            Player.computeActionReduction()
            Player.EffectList.append(GCDReductionEffect)
                                         # Computing AA delay
            Player.currentDelay = math.floor(math.floor(int(Player.baseDelay * 1000 ) * (100 - Player.Haste)/100)/10)/100

        fight_logging.debug("Starting simulation with TeamCompositionBonus = " + str(self.TeamCompositionBonus))
        fight_logging.debug("Parameters are -> RequirementOn : " + str(self.RequirementOn) + ", IgnoreMana : " + str(self.IgnoreMana))
        pB = ProgressBar.init(int(TimeLimit/TimeUnit), "Progress Of Fight (maxTime)", showBar=showProgress, extraBuffer=loadingBarBuffer)

        while(self.TimeStamp <= TimeLimit):

            for player in self.PlayerList:
                # if player.ActionSet[player.NextSpell] == None : player.TrueLock = True # Locking the player if None
                # Will first Check if the NextSpell is a GCD or not
                if(not player.TrueLock):# If it is we do nothing
                    if (player.ActionSet[player.NextSpell].GCD):
                        # Is a GCD
                        # Have to check if the player can cast the spell
                        # So check if Animation Lock, if Casting or if GCDLock
                        if(not (player.oGCDLock or player.GCDLock or player.Casting)):

                            player.CastingSpell = player.ActionSet[player.NextSpell].Cast(player, self.Enemy)# Cast the spell
                            # Locking the player
                            # print(Player.CastingSpell.CastTime)
                            # input(Player.CastingSpell.RecastTime)
                            player.Casting = True
                            player.CastingLockTimer = player.CastingSpell.CastTime
                            player.GCDLock = True
                            player.GCDLockTimer = player.CastingSpell.RecastTime
                            player.CastingTarget = self.Enemy
                            if player.CastingSpell.id > 0 and player.JobEnum != JobEnum.Pet:
                                log_str = ( "Timestamp : " + str(self.TimeStamp)
                                + " , Event : begin_cast_GCD"
                                + " , playerID : " + str(player.playerID)
                                + " , Ability : " + name_for_id(player.CastingSpell.id,player.ClassAction, player.JobAction) )

                                fight_logging.debug(log_str)

                        # Else we do nothing since doing the nextspell is not currently possible


                    else:
                        # Is an oGCD
                        # print("Spell with id : " + str(player.ActionSet[player.NextSpell].id))
                        # input("is being casted at : " + str(self.TimeStamp))
                        
                        if(not (player.oGCDLock or player.Casting)):
                            # Then we can cast the oGCD
                            player.CastingSpell = player.ActionSet[player.NextSpell].Cast(player, self.Enemy)
                            player.Casting = True
                            player.CastingLockTimer = player.CastingSpell.CastTime
                            player.CastingTarget = self.Enemy
                            player.oGCDLock = True
                            player.oGCDLockTimer = player.CastingSpell.CastTime

                            if player.CastingSpell.id > 0 and player.JobEnum != JobEnum.Pet:
                                log_str = ( "Timestamp : " + str(self.TimeStamp)
                                + " , Event : begin_cast_oGCD"
                                + " , playerID : " + str(player.playerID)
                                + " , Ability : " + name_for_id(player.CastingSpell.id,player.ClassAction, player.JobAction) )
                                
                                fight_logging.debug(log_str)
                

            if self.Enemy.hasEventList and start :
                # Only goes through the Enemy's EventList if the fight has started AND if the Enemy has an event list
                if not self.Enemy.IsCasting:
                    # If the enemy is not casting
                    self.Enemy.EventList[self.Enemy.EventNumber].begin_cast(self.Enemy) # Begins the casting of the next event

                self.Enemy.UpdateTimer(TimeUnit) # Updating the Enemy's timers

                        # We will now update any timer each player and the enemy has
            for player in self.PlayerList:
                player.updateTimer(TimeUnit)
                player.updateCD(TimeUnit)
                player.updateLock() # Update the lock on the player to see if the player's state changes
                                    # Castfinal is being called in player.updateLock() if it applies.

            # Updating shield timer and healing buff timer of all players
            for player in self.PlayerList:
                for shield in player.ShieldList: shield.UpdateTimer(TimeUnit) # Update shield timer
                for buff in player.ReceivedHealBuffList : buff.UpdateTimer(TimeUnit) # Update buff on received heal timer
                for buff in player.GivenHealBuffList : buff.UpdateTimer(TimeUnit) # Update buff on given heal timer
                for buff in player.MitBuffList : buff.UpdateTimer(TimeUnit) # Update Mit buff timer
                    
            for player in self.PlayerList:
                # Loops through the playerList
                # And calls every function in the player.EffectCDList
                # These functions will check if an effect should be terminated
                for CDCheck in player.EffectCDList:
                    CDCheck(player, self.Enemy)

                for remove in player.EffectToRemove:
                    # Loops through all effect that have been classified as terminated and removes them from the EffectCDList
                    player.EffectCDList.remove(remove) # Removing relevant spell
                    fight_logging.debug("Removing Check function : " + remove.__name__ + " TimeStamp : " + str(self.TimeStamp))
                for add in player.EffectToAdd:
                    # Adds any function to EffectCDList that should be added
                    player.EffectCDList.append(add)

                # Resets the list containing functions to be removed and added
                player.EffectToRemove = []
                player.EffectToAdd = []

            
            for player in self.PlayerList:
                             # Recomputing recastTime if new Haste has been added.
                             # Note that I am not certain if this goes here or in CastFinal.
                             # Having the recomputeRecastLock here means that as soon as a haste buff is applied
                             # it recomptes the recast lock based on the haste change. recomputeRecastLock()
                             # changes the aaDelay, current AATimer and the current recast lock but it does NOT
                             # recompute the current casting timer. This is because if you APPLY a haste buff it is either
                             # a GCD effect or an oGCD, meaning when you apply the casting timer should always be 0.
                             # And when the haste buff gets removed I BELIEVE that it does not affect the CURRENT casting
                             # but it will affect the recast. Hence why we do not recompute the casting timer (at least I think it works like that).
                if player.hasteHasChanged: 
                    player.recomputeRecastLock(isSpell=(player.RoleEnum == RoleEnum.Caster))
                for DOT in player.DOTList:
                    DOT.CheckDOT(player,self.Enemy, TimeUnit)


            if self.wipe: # If we detect that wipe has been set to true we stop the simulation. This for now only happens if a failedRequirement is fatal
                break


            for player in self.PlayerList:
                # Will go through all player and check if they have no more actions set to true. 
                # If so we will call the NextAction function
                if player.NoMoreAction:
                    # NextActionFunction is by default nothing and only returns True.
                    # But this function can be customized by the user to fit any use of it they might need
                    player.TrueLock = self.NextActionFunction(self, player)
                    if not player.TrueLock : 
                        player.NoMoreAction = False
                        player.NoMoreActionLog = True
                        log_str = "Player ID " + str(player.playerID) + " has received other actions to do, Timestamp : " + str(self.TimeStamp)
                        fight_logging.debug(log_str)
                    elif player.NoMoreActionLog:
                        player.NoMoreActionLog = False # We want this log to only happen once everytime the player has no more actions
                        log_str = "Player ID " + str(player.playerID) + " has no more actions, Timestamp : " + str(self.TimeStamp)
                        fight_logging.debug(log_str)


            CheckFinalLock = True
            for player in self.PlayerList:
                # Goes through every player and checks if they are done. If everyone has nothing to do the fight finishes
                CheckFinalLock = player.TrueLock and CheckFinalLock # If all player's TrueLock is true, then CheckFinalLock will be True
                

            if CheckFinalLock: 
                log_str = "Simulation has succesfully finished."
                fight_logging.debug(log_str)
                break
            
            
            if start:
                # If the fight has started, will sample DPS values at certain time.
                # The fight starts as soon as one player does damage.
                # The finished time is based on when the fight starts and not when the simulation starts.
                # If the simulation finishes before the fight starts there will be no damage done.
                if self.TimeStamp >= 1 and (isclose(self.TimeStamp%1, 0.25) or isclose(self.TimeStamp%1, 0.5) or isclose(self.TimeStamp%1, 0.75) or isclose(self.TimeStamp%1, 0)):# last thing is to ensure no division by zero and also to have no spike at the begining
                    # Samples DPS every frame of the simulation.
                    # If it becomes a problem if fights are too long, could limit rate of sampling.
                    self.timeValue+= [self.TimeStamp]
                    for Player in self.PlayerList:
                        Player.DPSGraph += [round(Player.TotalDamage/self.TimeStamp, 2)] # Rounding the value to 2 digits
                        Player.PotencyGraph += [round(Player.TotalPotency/self.TimeStamp, 2)]

            # Calling information extracting method
            self.ExtractInfo(self)

            # update self.TimeStamp
            self.TimeStamp += TimeUnit
            self.TimeStamp = round(self.TimeStamp,2) # Round it for cleaner value
            next(pB)

            if self.FightStart and not start:
                self.TimeStamp = 0
                start = True


        # Post fight computations

        if self.TimeStamp == 0 : self.TimeStamp = 1 # This is to fix any divide by zero error.

        remove = []

        for i in range(len(self.PlayerList)):  
            # Removing all instance of clones/summons from the fight. The DPS done has already been given to their master.
            player = self.PlayerList[i] 
            if player.JobEnum == JobEnum.Pet:
                remove += [i]

        k = 0
        for i in remove:
            self.PlayerList.pop(i-k)
            k+=1

        result_str = ""

        if self.wipe:
            # If the simulation ran into a problem caused by a failed requirement.
            result_str = "The simulation ran into a problem. You can disable requirements by setting 'Fight.RequirementOn' to 'False'.\n" + "=================\nList of failed fatal requirements : \n"

            for t in self.failedRequirementList: # Printing the failed requirement if it was fatal
                if t.fatal : 
                    result_str += (
                        "Failed Fatal Requirement : " + t.requirementName + " Timestamp : " + str(t.timeStamp) + "\n"
                        )
                             # Recording final HP value of players
        for gamer in self.PlayerList:
            gamer.HPGraph[0].append(self.TimeStamp)
            gamer.HPGraph[1].append(gamer.HP)

        # Printing the results if vocal is true.
        if self.TimeStamp < TimeLimit: pB.complete()
        fig2 = None
        fig = None
        if n > 0 and computeGraph : fig2 = SimulateRuns(self, n,showBar=showProgress, pBNewBuffer=loadingBarBuffer)
        if computeGraph : result, fig = PrintResult(self, self.TimeStamp, self.timeValue, PPSGraph=PPSGraph)
        else : result, fig = "", None
        if vocal:
            print(result_str + ("\n" if len(result_str) > 0 else "") + result) 
            if self.ShowGraph : 
                if fig != None:
                    try:
                        fig.show()
                        fig.savefig("DPSGraph.png")
                    except:
                        print("An error happened while trying to export the DPS Vs. Time graph. This could be due to a too large size.")
                if fig2 != None:
                    try:
                        fig2.show()
                        fig.savefig("DistGraph.png")
                    except:
                        print("Error while showing DPS distribution graph. This could be due to a too large size.")

        
        return result, fig, fig2
            
    def ComputeFunctions(self) -> None:
        """
        This function computes all relevant values needed to compute damage from potency using the stats of each player
        self : Fight -> Fight for which we want to compute the values (for all its players)
        """

        fight_logging.debug("Initializing damage values for all players.")

        for Player in self.PlayerList:
            levelMod = 1900
            baseMain = 390  
            baseSub = 400# Level 90 LevelMod values

            JobMod = Player.JobMod # Level 90 jobmod value, specific to each job

            Player.f_WD = (Player.Stat["WD"]+math.floor(baseMain*JobMod/1000))/100 
            Player.f_DET = math.floor(1000+math.floor(140*(Player.Stat["Det"]-baseMain)/levelMod))/1000# Determination damage
            if Player.RoleEnum == RoleEnum.Tank : Player.f_TEN = (1000+math.floor(100*(Player.Stat["Ten"]-baseSub)/levelMod))/1000 # Tenacity damage, 1 for non-tank player
            else : Player.f_TEN = 1 # if non-tank
            Player.f_SPD = (1000+math.floor(130*((Player.Stat["SS"] if Player.RoleEnum == RoleEnum.Caster or Player.RoleEnum == RoleEnum.Healer else Player.Stat["SkS"])-baseSub)/levelMod))/1000 # Used only for dots
            Player.CritRate = math.floor((200*(Player.Stat["Crit"]-baseSub)/levelMod+50))/1000 # Crit rate in decimal
            Player.CritMult = (math.floor(200*(Player.Stat["Crit"]-baseSub)/levelMod+400))/1000 # Crit Damage multiplier
            Player.DHRate = math.floor(550*(Player.Stat["DH"]-baseSub)/levelMod)/1000 # DH rate in decimal
            Player.DHAuto = math.floor(140*(Player.Stat["DH"]-baseSub)/levelMod)/1000 # DH bonus when auto crit/DH
                             # I was using baseMain for DHAuto, but I think that might be wrong based on what
                             # is in the tankcalc gear sheet. I have not yet confirmed which is the correct one to use.
            
            log_str = ("ID : " + str(Player.playerID) + " , Job : " + JobEnum.name_for_id(Player.JobEnum) 
            + " , f_WD : " + str(Player.f_WD) 
            + " , f_DET : " + str(Player.f_DET) 
            + " , f_TEN : " + str(Player.f_TEN) 
            + " , f_SPD : " + str(Player.f_SPD) 
            + " , f_CritRate : " + str(Player.CritRate) 
            + " , f_CritMult : " + str(Player.CritMult)
            + " , f_DHRate : " + str(Player.DHRate)  
            + ", f_DHAuto : " + str(Player.DHAuto))

            fight_logging.debug(log_str)

# HELPER FUNCTIONS UNDER

# GCDReduction Effect

def GCDReductionEffect(Player, Spell) -> None:
    """
    Computes the GCD reduction according to the SkillSpeed or the SpellSpeed of the player
    Player : player -> Player object
    Spell : Spell -> Spell object affected by the effect
    """
                             # Computation is done in computeActionTimer function.
                             # The spell object's property is also changed in this function.
    Player.computeActionTimer(Spell)

# Compute Damage
def ComputeDamage(Player, Potency, Enemy, SpellBonus, type, spellObj, SavePreBakedAction : bool = False, PlayerIDSavePreBakedAction : int = 1) -> float:

    """
    This function computes the damage from a given potency.
    Player : player -> player object doing the damage
    Potency : int -> potency value
    Enemy : Enemy -> Enemy object taking the damage
    SpellBonus : float -> Multiplying value to the final damage coming from the action itself
    type : int -> type of the action. 0 is Direct Damage, 1 is magical DOT, 2 is physical DOT and 3 is autos
    spellObj : Spell -> Object of the spell being casted
    SavePreBakedAction : bool -> If we want the simulator to save the PreBakedAction to the player's list. False by default.
    PlayerIDSavePreBakedAction : int -> ID of the player for which we want to record the PreBakedActions.
    """

                             # Creating page object to record action
    thisPage = page()
    thisPage.setPlayerID(Player.playerID)

    if type == 1 or type == 2:thisPage.setIsDOT()
    elif type ==3 : thisPage.setIsAuto()
    Player.CurrentFight.simulationRecord.addPage(thisPage)
    thisPage.setName(name_for_id(spellObj.id, Player.ClassAction, Player.JobAction))
    thisPage.setPotency(Potency)
    thisPage.setTimeStamp(Player.CurrentFight.TimeStamp)

    # Checking if under pot for record (only if not DOT)
    if Player.PotionTimer > 0 and (type != 1 and type != 2): thisPage.setHasPotion()

    # The type input signifies what type of damage we are dealing with, since the computation will chance according to what
    # type of damage it is

    # type = 0 (Direct Damage), type = 1 (magical DOT), type = 2(physical DOT), type = 3 (Auto-attacks)

    # All relevant formulas were taken from https://finalfantasy.fandom.com/wiki/Final_Fantasy_XIV_attributes#Damage_and_healing_formulae ,
    # were given to me by javaJake#0001 on discord or were taken from the Allagan Studies discord server.
    # The formulas on the website assume a random function that will randomise the ouput. We instead compute the expected outcome.
    # Also thanks to whoever did the DPS computation code on the black mage gear comparison sheet : https://docs.google.com/spreadsheets/d/1t3EYSOPuMceqCFrU4WAbzSd4gbYi-J7YeMB36dNmaWM/edit# gid=654212594
    # It helped me a lot to understand better the DPS computation of this game
    # Also, note that this function is still in development, and so some of these formulas might be a bit off. Use at your own risk.
    # This function will compute the DPS given the stats of a player

                # Limit Break damage logic go here
                # TODO
    if spellObj.type == 3: return 0, 1000


    # These computations should be up to date with Endwalker.
    baseMain = 390  
    Enemy = Player.CurrentFight.Enemy # Enemy targetted
    isPet = (Player.JobEnum == JobEnum.Pet)
    isTank = Player.RoleEnum == RoleEnum.Tank or (isPet and Player.Master.RoleEnum == RoleEnum.Tank)
    MainStat = Player.Stat["MainStat"]
    # Computing values used throughout all computations
    if isTank : f_MAIN_DMG = (100+math.floor((MainStat-baseMain)*156/baseMain))/100 # Tanks have a difference constant 
    else: f_MAIN_DMG = (100+math.floor((MainStat-baseMain)*195/baseMain))/100
    # These values are all already computed since they do not change
    f_WD = Player.f_WD
    f_DET = Player.f_DET
    f_TEN = Player.f_TEN
    f_SPD = Player.f_SPD
    CritRate = (Player.CritRate)
    CritMult = Player.CritMult
    DHRate = Player.DHRate
    CritRateBonus = Player.CritRateBonus
    DHRateBonus = Player.DHRateBonus # Saving value for later use if necessary
    f_DET_DH = math.floor((f_DET + Player.DHAuto) * 1000 ) / 1000

                             # Check if DOT. If not we take current buffs.
                             # If is a dot we take snapshotted buffs.
    if (type == 0 or type == 3) or not spellObj.onceThroughFlag:
        if round(Player.CritRateBonus,2) > 0 : thisPage.addCritBuffList(("Other", Player.CritRateBonus))
        if round(Player.DHRateBonus,2) > 0 : thisPage.addDHBuffList(("Other", Player.DHRateBonus))

        if Enemy.ChainStratagem: 
            CritRateBonus += 0.1    # If ChainStratagem is active, increase crit rate
            thisPage.addCritBuffList(("CS", 0.1))

        if Enemy.WanderingMinuet: 
            CritRateBonus += 0.02 # If WanderingMinuet is active, increase crit rate
            thisPage.addCritBuffList(("WM", 0.02))

        if Enemy.ArmyPaeon: 
            DHRateBonus += 0.03 # If WanderingMinuet is active, increase crit rate
            thisPage.addDHBuffList(("AP", 0.03))

        if Enemy.BattleVoice: 
            DHRateBonus += 0.2 # If BattleVoice is active, increase DHRate
            thisPage.addDHBuffList(("BV", 0.2))
    else:
                             # If dot and has gone through once (so has snapshotted buff) we use them
                             # This only considers Crit,DH and Pot snapshot
        DHRateBonus = spellObj.DHBonus
        if round(DHRateBonus, 2) > 0 : thisPage.addDHBuffList(("DOTSnapshot", DHRateBonus))
        CritRateBonus = spellObj.CritBonus
        if round(CritRateBonus, 2) > 0 : thisPage.addCritBuffList(("DOTSnapshot", CritRateBonus))

                             # Checking if ChainStratagem clips since ground DOT do not clip
                             # Debuff
        if spellObj.isGround and Enemy.ChainStratagem:
            CritRateBonus += 0.1    # If ChainStratagem is active, increase crit rate
            thisPage.addCritBuffList(("CS", 0.1))
        
        if spellObj.potSnapshot : 
                             # We have to recompute f_AP.
            if Player.PotionTimer <= 0 :
                             # If a potion is already applied f_MAIN_DMG won't change so we skip
                             # So we only change f_MAIN_DMG when the potion is no longer active
                             # on the player but the DOT has it snapshotted.
                mainStatBonus = min(int(MainStat * 0.1), 262)
                if isTank : f_MAIN_DMG = (100+math.floor(((MainStat+mainStatBonus)-baseMain)*156/baseMain))/100 # Tanks have a difference constant 
                else: f_MAIN_DMG = (100+math.floor(((MainStat+mainStatBonus)-baseMain)*195/baseMain))/100
            thisPage.setHasPotion()
        elif not spellObj.potSnapshot and Player.PotionTimer > 0:
                             # If not snapshot Potion BUT potion is currently applied
                             # We have to recalculate f_MAIN_DMG using the lower
                             # MainStatvalue
            if isTank : f_MAIN_DMG = (100+math.floor(((MainStat-Player.mainStatBonus)-baseMain)*156/baseMain))/100 # Tanks have a difference constant 
            else: f_MAIN_DMG = (100+math.floor(((MainStat-Player.mainStatBonus)-baseMain)*195/baseMain))/100
    DHRate += DHRateBonus# Adding Bonus
    CritRate += CritRateBonus# Adding bonus

    # We will check if the ability is an assured crit and/ord DH, in which case we will have to buff the damage
    # Depending on the buffs the player is currently receiving

    auto_crit = False
    auto_DH = False

    if type == 0: # Making sure its not an AA or DOT
        if Player.JobEnum == JobEnum.Machinist: 
            # Then if machinist, has to check if direct crit guarantee
            if Player.ActionSet[Player.NextSpell].id != -1 and Player.ActionSet[Player.NextSpell].id != -2 and Player.Reassemble and Player.ActionSet[Player.NextSpell].Weaponskill:    # Checks if reassemble is on and if its a weapon skill
                CritRate = 1
                DHRate = 1
                Player.Reassemble = False # Uses Reassemble    
                auto_crit = True
                auto_DH = True   
        elif Player.JobEnum == JobEnum.Warrior:
            fight_logging.debug("Inner Release Stack : " + str(Player.InnerReleaseStack))        # Primal Rend                                     # Fell Cleave                                    # Inner Chaos
            if (Player.NextSpell < len(Player.ActionSet)) and (Player.ActionSet[Player.NextSpell].id == 25753 or (Player.ActionSet[Player.NextSpell].id == 3549 and Player.InnerReleaseStack > 0) or Player.ActionSet[Player.NextSpell].id == 16465):
                CritRate = 1# If inner release weaponskill
                DHRate = 1
                             # Remove stack if fell cleave
                if Player.ActionSet[Player.NextSpell].id == 3549 : Player.InnerReleaseStack -= 1
                auto_crit = True
                auto_DH = True
        elif Player.JobEnum == JobEnum.Samurai:
            if Player.DirectCrit:
                CritRate = 1
                DHRate = 1
                Player.DirectCrit = False
                auto_crit = True
                auto_DH = True
        elif Player.JobEnum == JobEnum.Dancer:
            if Player.NextDirectCrit:
                CritRate = 1
                DHRate = 1
                Player.NextDirectCrit = False
                auto_crit = True
                auto_DH = True
        elif Player.JobEnum == JobEnum.Dragoon:
            if Player.NextCrit and Player.ActionSet[Player.NextSpell].Weaponskill: # If next crit and weaponskill
                CritRate = 1
                Player.NextCrit = False
                auto_crit = True
        elif Player.JobEnum == JobEnum.Monk:
            if Player.GuaranteedCrit and Player.ActionSet[Player.NextSpell].Weaponskill:
                CritRate = 1
                Player.GuaranteedCrit = False
                auto_crit = True

    if type == 0: fight_logging.debug(str((Player.Stat["MainStat"],f_MAIN_DMG, f_WD, f_DET, f_TEN, f_SPD, CritRate, CritMult, DHRate)))


    if type == 0: # Type 0 is direct damage
        #Damage = Potency * f_MAIN_DMG * f_DET * f_TEN  *f_WD * Player.Trait
        Damage = math.floor(math.floor(math.floor(math.floor(math.floor(Potency * f_MAIN_DMG) * (f_DET_DH if auto_crit and auto_DH else f_DET)) * f_TEN ) *f_WD) * Player.Trait) # Player.Trait is trait DPS bonus
        Damage = math.floor(Damage * SpellBonus)
        Player.NumberDamageSpell += 1
        Player.CritRateHistory += [CritRate]
    elif type == 1 : # Type 1 is magical DOT
        Damage = math.floor(math.floor(math.floor(math.floor(math.floor(math.floor(Potency * f_WD) * f_MAIN_DMG) * f_SPD) * f_DET) * f_TEN) * Player.Trait) + 1
        
        if not spellObj.onceThroughFlag:# If we haven't gotten through with this DOT once, we have to snapshot the buffs


            spellObj.DHBonus = DHRateBonus # Adding Bonus
                             # Adding Crit bonus but removing ChainStratagem if groundDOT since
                             # Chain Strat is a debuff.
            spellObj.CritBonus = CritRateBonus - (0.1 if spellObj.isGround and Enemy.ChainStratagem else 0)
            spellObj.potSnapshot = Player.PotionTimer > 0 # Checking Potion

            for buffs in Player.buffList: 
                             # Adding buff to DOT only if not (ground and debuff) since ground DOT
                             # do not snapshot debuff.
                if not (spellObj.isGround and buffs.isDebuff) : spellObj.MultBonus += [buffs] 
            for buffs in Enemy.buffList:
                             # Adding buff to DOT only if not (ground and debuff) since ground DOT
                             # do not snapshot debuff.
                if not (spellObj.isGround and buffs.isDebuff) : spellObj.MultBonus += [buffs] 

            # Now the DOT has completely snapshot all possible buff. So we save those
            # and never come back here

            spellObj.onceThroughFlag = True # set flag to True, so never snapshot again

    elif type == 2: # Physical DOT
        Damage = math.floor(math.floor(math.floor(math.floor(math.floor(math.floor(Potency * f_MAIN_DMG) * f_DET) * f_TEN) * f_SPD) * f_WD) * Player.Trait) +1
    
        if not spellObj.onceThroughFlag:# If we haven't gotten through with this DOT once, we have to snapshot the buffs

            spellObj.DHBonus = DHRateBonus # Adding DH Bonus
                             # Adding Crit bonus but removing ChainStratagem if groundDOT since
                             # Chain Strat is a debuff.
            spellObj.CritBonus = CritRateBonus - (0.1 if spellObj.isGround and Enemy.ChainStratagem else 0)
            spellObj.potSnapshot = Player.PotionTimer > 0 # Checking Potion

            for buffs in Player.buffList: 
                             # Adding buff to DOT only if not (ground and debuff) since ground DOT
                             # do not snapshot debuff.
                if not (spellObj.isGround and buffs.isDebuff) : spellObj.MultBonus += [buffs] 
            for buffs in Enemy.buffList:
                             # Adding buff to DOT only if not (ground and debuff) since ground DOT
                             # do not snapshot debuff.
                if not (spellObj.isGround and buffs.isDebuff) : spellObj.MultBonus += [buffs] 

            # Now the DOT has completely snapshot all possible buff. So we save those
            # and never come back here

            spellObj.onceThroughFlag = True # set flag to True, so never snapshot again

    elif type == 3: # Auto-attacks
        Damage = math.floor(math.floor(math.floor(math.floor(Potency * f_MAIN_DMG) * f_DET) * f_TEN) * f_SPD)
        Damage = math.floor(math.floor(Damage * math.floor(f_WD * (Player.baseDelay/3) *100 )/100) * Player.Trait)
    # Now applying buffs

    if type == 0 or type == 3 or not spellObj.onceThroughFlag: # If Action or AA, then we apply the current buffs
        for buffs in Player.buffList: 
            Damage = math.floor(Damage * buffs.MultDPS) # Multiplying all buffs
            thisPage.addPercentBuff(buffs)
        for buffs in Enemy.buffList:
            Damage = math.floor(Damage * buffs.MultDPS) # Multiplying all buffs
            thisPage.addPercentBuff(buffs)
    else: # if type is 1 or 2, then its a DOT, so we have to use the snapshotted buffs
        for buffs in spellObj.MultBonus:
            Damage = math.floor(Damage * buffs.MultDPS)
            thisPage.addPercentBuff(buffs)
                             # If DOT is a ground DOT we have to add buffs that are debuff
        if spellObj.isGround:
            for buffs in Player.buffList: 
                if buffs.isDebuff:
                    Damage = math.floor(Damage * buffs.MultDPS) # Multiplying all debuffs
                    thisPage.addPercentBuff(buffs)
            for buffs in Enemy.buffList:
                if buffs.isDebuff:
                    Damage = math.floor(Damage * buffs.MultDPS) # Multiplying all debuffs
                    thisPage.addPercentBuff(buffs)


    if SavePreBakedAction and (Player.playerID == PlayerIDSavePreBakedAction or (isPet and Player.Master.playerID == PlayerIDSavePreBakedAction)): 
        """
        If that is set to true we will record all we need and will not compute the rest.
        We will check if the action is a GCD with recast time of lesser or equal to 1.5s since the GCD
        cannot go lower. The total time will be remembered and substracted from the total time that is reduceable from more SpS.
        """
                             # Snapshotting all buff on that action.
        buffList = []
        potActive = False
        if type == 0 or type == 3 or not spellObj.onceThroughFlag:
                             # if AA or action take all curent buff/debuff
            for buff in Player.buffList:
                buffList.append(buff)
            for buff in Enemy.buffList:
                buffList.append(buff)

            potActive = Player.PotionTimer > 0 or (isPet and Player.Master.PotionTimer > 0)

        else:
                             # If is DOT we take the snapshotted buffs + debuff if ground
            for buff in spellObj.MultBonus:
                buffList.append(buff)
                             # If DOT is a ground DOT we have to add buffs that are debuff
            if spellObj.isGround:
                for buff in Player.buffList: 
                    if buff.isDebuff: buffList.append(buff)
                for buff in Enemy.buffList:
                    if buff.isDebuff: buffList.append(buff)

            potActive = spellObj.potSnapshot

        if auto_crit and auto_DH : fight_logging.debug("Auto Crit/DH prebaked")
        elif auto_crit : fight_logging.debug("Auto Crit prebaked")

        nonReducableStamp = 0 if not Player.CurrentFight.FightStart else Player.totalTimeNoFaster
        reducableStamp = 0 if not Player.CurrentFight.FightStart else Player.CurrentFight.TimeStamp - Player.totalTimeNoFaster

        gcdLockTimer = max(spellObj.notRoundRecastTime, spellObj.notRoundCastTime) if spellObj.GCD  and (Player.RoleEnum != RoleEnum.Pet) and max(spellObj.notRoundRecastTime, spellObj.notRoundCastTime) > 1.5 else 0
        if spellObj.GCD and (Player.RoleEnum == RoleEnum.Melee or Player.RoleEnum == RoleEnum.Tank) and spellObj.type == 1 : gcdLockTimer = 0
        elif spellObj.GCD and (Player.RoleEnum == RoleEnum.Caster) and spellObj.type == 2 : gcdLockTimer = 0
        
        newPreBaked = PreBakedAction(isTank, Player.CurrentFight.TeamCompositionBonus,buffList, Player.Trait, Potency, type,Player.CurrentFight.TimeStamp if Player.CurrentFight.FightStart else 0, nonReducableStamp + (0 if type == 0 else reducableStamp), 
                                     reducableStamp if type == 0 else 0 ,AutoCrit=auto_crit, AutoDH=auto_DH, isFromPet=isPet, isGCD=spellObj.GCD,gcdLockTimer=gcdLockTimer,spellDPSBuff=SpellBonus, isConditionalAction=spellObj.conditionalAction)
                             # Giving dh and crit bonus
        newPreBaked.CritBonus = CritRateBonus
        newPreBaked.DHBonus = DHRateBonus
        newPreBaked.potionIsActive = potActive
                             # have to specific if is wildfire
        if spellObj.id == -2878 : newPreBaked.isWildfire = True

        (Player if not isPet else Player.Master).PreBakedActionSet.append(newPreBaked)
        
        return Potency, Potency        # Exit the function since we are not interested in the immediate damage value. Still return potency as to not break the fight's duration.


    if spellObj.id == -2878: #If wildfire it cannot crit or DH, so we remove it
        non_crit_dh_expected, dh_crit_expected = Damage, Damage # Non crit expected damage, expected damage with crit
        (Player if Player.JobEnum != JobEnum.Pet else Player.Master).ZIPActionSet.append(ZIPAction(Damage, 0, CritMult, 0))
        thisPage.setDamage(dh_crit_expected)
        return non_crit_dh_expected , dh_crit_expected

    
    if auto_crit and auto_DH: # If both 
        fight_logging.debug("Auto Crit/DH")
        auto_crit_bonus = (1 + CritRateBonus * CritMult) # Auto_crit bonus if buffed
        fight_logging.debug("CritRate : " + str(CritRateBonus) + " autocritbonus : " + str(auto_crit_bonus))
        auto_dh_bonus = (1 + DHRateBonus * 0.25) # Auto_DH bonus if buffed
        fight_logging.debug("DHRateBonus : " + str(DHRateBonus) + " autodhbonus : " + str(auto_dh_bonus))
        non_crit_dh_expected, dh_crit_expected =  0, math.floor(math.floor(Damage * (1 + CritMult) ) * (1.25)) 
        (Player if Player.JobEnum != JobEnum.Pet else Player.Master).ZIPActionSet.append(ZIPAction(Damage, 1, CritMult, 1, auto_crit=True, auto_dh=True, AutoCritBonus=auto_crit_bonus, AutoDHBonus=auto_dh_bonus))
        Damage = math.floor(math.floor(dh_crit_expected * auto_crit_bonus) * auto_dh_bonus)
        thisPage.setDamage(Damage)
        thisPage.setAutoCrit(True)
        thisPage.setAutoDH(True)
        return 0, Damage
    elif auto_crit: # If sure to crit, add crit to min expected damage
        fight_logging.debug("Auto Crit")
        auto_crit_bonus = (1 + CritRateBonus * CritMult) # Auto_crit bonus if buffed
        non_crit_dh_expected, dh_crit_expected = ( 0, 
                                                   math.floor(math.floor(Damage * (1 + CritMult) ) * (1 + (DHRate * 0.25))) )# If we have auto crit, we return full damage
        (Player if Player.JobEnum != JobEnum.Pet else Player.Master).ZIPActionSet.append(ZIPAction(Damage, 1, CritMult, DHRate, auto_crit=True, AutoCritBonus=auto_crit_bonus ))
        Damage = math.floor(dh_crit_expected * auto_crit_bonus) 
        thisPage.setDamage(Damage)
        thisPage.setAutoCrit(True)
        return 0, Damage
    else:# No auto_crit or auto_DH
        non_crit_dh_expected, dh_crit_expected = 0, math.floor(math.floor(Damage * (1 + (CritRate * CritMult)) ) * (1 + (DHRate * 0.25))) # Non crit expected damage, expected damage with crit
        (Player if Player.JobEnum != JobEnum.Pet else Player.Master).ZIPActionSet.append(ZIPAction(Damage, CritRate, CritMult, DHRate))
        Damage = dh_crit_expected
        thisPage.setDamage(Damage)
        return 0 , Damage

# Compute Healing
def ComputeHeal(Player, Potency, Target, SpellBonus, type, spellObj) -> float:
    """This function computes and returns the healing done by an action.

    Args:
        Player (Player): Player casting the action
        Potency (int): Potency of the heal
        Target (Player): Target of the healing
        SpellBonus (float): Bonus of the spell
        type (int): Type of the action. type = 1 is Direct Heal, Type = 2 is DOT heal
        spellObj (Spell): Object corresponding to the action being casted
    """
    baseMain = 390  


    if Player.JobEnum == JobEnum.Pet: MainStat = Player.Stat["MainStat"] # Summons do not receive bonus
    else: MainStat = math.floor(Player.Stat["MainStat"] * Player.CurrentFight.TeamCompositionBonus) # Scaling %bonus on mainstat
    # Computing values used throughout all computations
    f_MAIN_heal = (100+math.floor((MainStat-baseMain)*304/baseMain))/100
    # These values are all already computed since they do not change
    f_WD = Player.f_WD
    f_DET = Player.f_DET
    f_TEN = Player.f_TEN
    f_SPD = Player.f_SPD
    CritRate = Player.CritRate
    CritMult = Player.CritMult
    
    # WARNING
    # THESE CONSTANTS ARE PROBABLY WRONG. I HAVE NOT VERIFIED AND AM SIMPLY USING THOSE IN ORDER TO TEST THE CODE.
    
    if type == 1: # Direct Heal
        H_1 = math.floor(math.floor(math.floor(math.floor(math.floor(Potency * f_MAIN_heal * f_DET) * f_DET) * f_TEN) * f_WD ) * Player.Trait)
        H_1_min = math.floor(H_1 * 97/100) # Minimal healing done
        H_1_expected_crit = math.floor(H_1 * (1 + round((CritRate * CritMult), 3))) # Expected healing done
        # All buff the Player is giving
        for HealingBuff in Player.GivenHealBuffList:
            H_1_min = math.floor(H_1_min * HealingBuff.PercentBuff)
            H_1_expected_crit = math.floor(H_1_expected_crit * HealingBuff.PercentBuff)
        # All buff the target is receiving
        for HealingBuff in Target.ReceivedHealBuffList:
            H_1_min = math.floor(H_1_min * HealingBuff.PercentBuff)
            H_1_expected_crit = math.floor(H_1_expected_crit * HealingBuff.PercentBuff)

        return H_1_min, H_1_expected_crit 
    if type == 2: # DOT heal
        H_1 = math.floor(math.floor(math.floor(math.floor(math.floor(math.floor(Potency * f_MAIN_heal * f_DET) * f_DET) * f_TEN)  * f_SPD)* f_WD ) * Player.Trait)
        H_1_min = math.floor(H_1 * 97/100) # Minimal healing done
        H_1_expected_crit = math.floor(H_1 * (1 + round((CritRate * CritMult), 3))) # Expected healing done

        # If this is the first time the DOT is used. Will snapshot all buffs for later application of the heal.

        if not spellObj.onceThroughFlag:
            spellObj.onceThroughFlag = True
            # If this DOT has never been through. We take note of what buffs
            for HealingBuff in Player.GivenHealBuffList:
                spellObj.MultBonus.append(HealingBuff.PercentBuff)
            # All buff the target is receiving
            for HealingBuff in Target.ReceivedHealBuffList:
                spellObj.MultBonus.append(HealingBuff.PercentBuff)

        for buff in spellObj.MultBonus:
            # Applying all buff saved. buff here is the percent bonus in float
            H_1_min = math.floor(H_1_min * buff)
            H_1_expected_crit = math.floor(H_1_expected_crit * buff)

        return H_1_min, H_1_expected_crit



# -- coding: utf-8 --
"""
THIS CODE IS OUT OF DATE. USE FFLogs_api.py INSTEAD. IMPORTING THIS MODULE WILL RESULT IN AN ERROR.

@author: Bri

The network/API request part of this code was made by a friend of mine called Brian.
I only did the code relevant to how we used the data, he did everything regarding
how to get the data. You can DM him on discord if you have questions : Bri-kun#6539

"""
from ffxivcalc.helperCode.exceptions import outDatedFFLogsCode
raise(outDatedFFLogsCode)


from ffxivcalc.Jobs.Base_Spell import PrepullPotion, WaitAbility
from ffxivcalc.Jobs.Player import Player
from ffxivcalc.Jobs.PlayerEnum import *
from ffxivcalc.Jobs.ActionEnum import name_for_id
#CASTER

from ffxivcalc.Jobs.Caster.Caster_Spell import CasterAbility
from ffxivcalc.Jobs.Caster.Blackmage.BlackMage_Spell import BlackMageAbility
from ffxivcalc.Jobs.Caster.Redmage.Redmage_Spell import RedMageAbility
from ffxivcalc.Jobs.Caster.Summoner.Summoner_Spell import SummonerAbility

#HEALER
from ffxivcalc.Jobs.Healer.Healer_Spell import HealerAbility
from ffxivcalc.Jobs.Healer.Sage.Sage_Spell import SageAbility
from ffxivcalc.Jobs.Healer.Astrologian.Astrologian_Spell import AstrologianAbility
from ffxivcalc.Jobs.Healer.Scholar.Scholar_Spell import ScholarAbility
from ffxivcalc.Jobs.Healer.Whitemage.Whitemage_Spell import WhiteMageAbility
from ffxivcalc.Jobs.Melee.Monk.Monk_Spell import MonkAbility

#RANGED
from ffxivcalc.Jobs.Ranged.Ranged_Spell import BardSpell, RangedAbility
from ffxivcalc.Jobs.Ranged.Bard.Bard_Spell import BardAbility
from ffxivcalc.Jobs.Ranged.Machinist.Machinist_Spell import MachinistAbility
from ffxivcalc.Jobs.Ranged.Dancer.Dancer_Spell import DancerAbility

#TANK
from ffxivcalc.Jobs.Tank.Tank_Spell import TankAbility
from ffxivcalc.Jobs.Tank.Gunbreaker.Gunbreaker_Spell import GunbreakerAbility
from ffxivcalc.Jobs.Tank.DarkKnight.DarkKnight_Spell import DarkKnightAbility
from ffxivcalc.Jobs.Tank.Warrior.Warrior_Spell import WarriorAbility
from ffxivcalc.Jobs.Tank.Paladin.Paladin_Spell import PaladinAbility

#MELEE
from ffxivcalc.Jobs.Melee.Melee_Spell import MeleeAbility
from ffxivcalc.Jobs.Melee.Samurai.Samurai_Spell import MeikyoCheck, MeikyoEffect, MeikyoStackCheck, SamuraiAbility
from ffxivcalc.Jobs.Melee.Ninja.Ninja_Spell import ApplyHuton, NinjaAbility
from ffxivcalc.Jobs.Melee.Dragoon.Dragoon_Spell import DragoonAbility
from ffxivcalc.Jobs.Melee.Reaper.Reaper_Spell import ReaperAbility

from ffxivcalc.helperCode.exceptions import ActionNotFound, JobNotFound, InvalidTarget

import http.client, json, logging
main_logging = logging.getLogger("ffxivcalc")
fflogsapi_logging = main_logging.getChild("FFLogsAPI")

from ffxivcalc.Jobs.Caster.Blackmage.BlackMage_Spell import ElementalEffect
from ffxivcalc.Jobs.Caster.Redmage.Redmage_Spell import DualCastEffect
from ffxivcalc.Jobs.Melee.Monk.Monk_Spell import ComboEffect
from ffxivcalc.Jobs.Ranged.Bard.Bard_Spell import SongEffect
from ffxivcalc.Jobs.Ranged.Dancer.Dancer_Spell import EspritEffect
from ffxivcalc.Jobs.Tank.Warrior.Warrior_Spell import SurgingTempestEffect 

class ActionNotFound(Exception):#Exception called if an action isn't found in the dictionnary
    pass
class JobNotFound(Exception):#Exception called if a Job isn't found
    pass


def lookup_abilityID(actionID, targetID, sourceID, player_list):
    """
    This function will translate an actionID into a Spell object of the relevant action that the simulator can use.
    actionID : int -> ID of the action in the game
    targetID : int -> ID of the target. Can be player or Enemy.
    sourceID : int -> ID of the player casting the action.
    player_list : dict -> dict of all players with some information.
    """
    #Will first get the job of the sourceID so we know in what dictionnary to search for

    def lookup(JobDict, ClassDict, job_name):
        """
        This function actually looks up the relevant dictionnary of the Job to find the Spell object.
        JobDict : dict -> dictionnary with keys being IDs and mapping to the Spell object (only for Job actions)
        ClassDict : dict -> same as JobDict, but for Class actions
        """
        if not (int(actionID) in JobDict.keys()): #if not in, then the action is in the ClassDict
            if not (int(actionID) in ClassDict.keys()):
                log_str = "Action Not found , Job : " + job_name + " , ActionId : " + str(actionID) + " , targetID : " + str(targetID) + " , sourceID : " + str(sourceID)
                fflogsapi_logging.warning(log_str)
                fflogsapi_logging.warning("Since action was not found defaulting to WaitAbility(0).")
                return WaitAbility(0) #Currently at none so we can debug
                raise ActionNotFound #Did not find action
            if callable(ClassDict[int(actionID)]): #If the action is a function
                if (not (str(targetID) in player_list.keys())):
                    player_obj = player_list[sourceID]["job_object"]
                    raise InvalidTarget(name_for_id(actionID, player_obj.ClassAction, player_obj.JobAction), player_obj, None,True, targetID)
                return ClassDict[int(actionID)](player_list[str(targetID)]["job_object"])
            return ClassDict[int(actionID)] #Class actions do not have the possibility to target other allies, so we assume itll target an enemy

        if callable(JobDict[int(actionID)]): #If the action is a function
            if (not (str(targetID) in player_list.keys())):
                player_obj = player_list[sourceID]["job_object"]
                raise InvalidTarget(name_for_id(actionID, player_obj.ClassAction, player_obj.JobAction), player_obj, None,True, targetID)
            return JobDict[int(actionID)](player_list[str(targetID)]["job_object"])
        return JobDict[int(actionID)] #Else return object

    job_name = player_list[str(sourceID)]["job"] #getting job name

    #Will now go through all possible job and find what action is being used based on the ID. If the ID is not right, it will
    #raise an ActionNotFoundError. And if the job's name does not exist it will raise a JobNotFoundError
    if job_name == "BlackMage" :#Caster
        return lookup(BlackMageAbility, CasterAbility,job_name)
    elif job_name == "RedMage":
        return lookup(RedMageAbility, CasterAbility,job_name)
    elif job_name == "Summoner":
        return lookup(SummonerAbility, CasterAbility,job_name)
    elif job_name == "Dancer":#Ranged
        return lookup(DancerAbility, RangedAbility,job_name)
    elif job_name == "Machinist":
        return lookup(MachinistAbility, RangedAbility,job_name)
    elif job_name == "Bard":
        return lookup(BardAbility, RangedAbility,job_name)
    elif job_name == "Warrior":#Tank
        return lookup(WarriorAbility, TankAbility,job_name)
    elif job_name == "Gunbreaker":
        return lookup(GunbreakerAbility, TankAbility,job_name)
    elif job_name == "DarkKnight":
        return lookup(DarkKnightAbility, TankAbility,job_name)
    elif job_name == "Paladin":
        return lookup(PaladinAbility, TankAbility,job_name)
    elif job_name == "WhiteMage":#Healer
        return lookup(WhiteMageAbility, HealerAbility,job_name)
    elif job_name == "Scholar":
        return lookup(ScholarAbility, HealerAbility,job_name)
    elif job_name == "Sage":
        return lookup(SageAbility, HealerAbility,job_name)
    elif job_name == "Astrologian":
        return lookup(AstrologianAbility, HealerAbility,job_name)
    elif job_name == "Samurai":#Melee
        return lookup(SamuraiAbility, MeleeAbility,job_name)
    elif job_name == "Reaper":
        return lookup(ReaperAbility, MeleeAbility,job_name)
    elif job_name == "Ninja":
        return lookup(NinjaAbility, MeleeAbility,job_name)
    elif job_name == "Monk":
        return lookup(MonkAbility, MeleeAbility,job_name)
    elif job_name == "Dragoon":
        return lookup(DragoonAbility, MeleeAbility,job_name)

    fflogsapi_logging.critical("Job name not found : " + job_name)

    raise JobNotFound #If we get here, then we have not found the job in question
    #This should not happen, and if it does it means we either have a serious problem or the names aren't correct

def getAccessToken(conn, client_id, client_secret):
    """
    This is called upon the execution and returns an acces_token to request data from FFLogs
    """
    payload = "grant_type=client_credentials&client_id=%s&client_secret=%s" % (client_id, client_secret)
    headers = {'content-type':"application/x-www-form-urlencoded"}
    conn.request("POST","/oauth/token", payload, headers)
    res = conn.getresponse()
    res_str = res.read().decode("utf-8")
    res_json = json.loads(res_str)
    return res_json["access_token"]

def getAbilityList(fightID, fightNumber):

    """
    This function  takes information of a FFLog report and transform it into a list of spell object the simulator can read
    fightID : int -> ID of the report
    fightNmber : int -> ID of the fight in the report
    """

    client_id = "9686da23-55d6-4f64-bd9d-40e2c64f8edf" #Put your own client_id and client_secret obtained from FFLogs
    client_secret = "ioZontZKcMxZwc33K4zsWlMAPY5dfZKsuo3eSFXE" #Supposed to be secret >.>
    fflogsapi_logging.debug("Sending request to fflogs, client_id : " + client_id + " , client_secret : " + client_secret)
    conn = http.client.HTTPSConnection("www.fflogs.com")
    access_token = getAccessToken(conn, client_id, client_secret)

    payload = "{\"query\":\"query trio{\\n\\treportData {\\n\\t\\treport(code: \\\"" + fightID + "\\\") {\\n\\t\\t\\tplayerDetails(fightIDs:" + fightNumber +",endTime:999999999),\\n\\t\\t\\tfights(fightIDs:" + fightNumber +"){\\n\\t\\t\\t\\tenemyNPCs{\\n\\t\\t\\t\\t\\tid\\n\\t\\t\\t\\t}\\n\\t\\t\\t\\tstartTime\\n\\t\\t\\t}\\n\\t\\t}\\n\\t\\t\\t\\n\\t}\\n}\",\"operationName\":\"trio\"}"

    headers = {
        'Content-Type': "application/json",
        'Authorization': "Bearer %s" % access_token
        }
    conn.request("POST", "/api/v2/client", payload, headers)

    res = conn.getresponse()
    data = res.read()
    data_json = json.loads(data.decode("utf-8"))\

    #Getting Player's Class, ids and name
    player_data = data_json["data"]["reportData"]["report"]["playerDetails"]["data"]["playerDetails"]
    enemy_data = data_json["data"]["reportData"]["report"]["fights"][0]["enemyNPCs"]
    #Mix of dictionnary and array with relevant information
    #The goal of this request and the following code is to parse the JSON file into relevant information
    player_list = {} #Dict which will have all a list of players with their ids, role and name
    enemy_list = [] #List with all ids of enemies in the fight
    relative_timestamp_zero = int(data_json["data"]["reportData"]["report"]["fights"][0]["startTime"]) #relative 0 of the report
    fflogsapi_logging.debug("Setting up enemy ids")
    for enemy in enemy_data:
        enemy_list += [enemy["id"]] #Getting the enemies id so we can identifiate who an enemy is not targeted by an action

    for player_class in player_data: #player_data is a dictionnary with key "healers", "DPS", "tanks"
        for player in player_data[player_class]:
            fflogsapi_logging.debug("Creating Player object for playerID : " + str(player["id"]))
            #Will check what job the player is so we can create a player object of the relevant job

            job_name = player["type"]
            job_object = None

            if job_name == "Sage" : job_object = Player([], [],{}, JobEnum.Sage)
            elif job_name == "Scholar" : job_object = Player([], [],{}, JobEnum.Scholar)
            elif job_name == "WhiteMage" : job_object = Player([], [],{}, JobEnum.WhiteMage)
            elif job_name == "Astrologian" : job_object = Player([], [],{}, JobEnum.Astrologian)
            #Tank
            elif job_name == "Warrior" : job_object = Player([], [SurgingTempestEffect],{}, JobEnum.Astrologian)
            elif job_name == "DarkKnight" : job_object = Player([], [],{}, JobEnum.DarkKnight)
            elif job_name == "Paladin" : job_object = Player([], [],{}, JobEnum.Paladin)
            elif job_name == "Gunbreaker" : job_object = Player([], [],{}, JobEnum.Gunbreaker)
            #Caster
            elif job_name == "BlackMage" : job_object = Player([], [],{}, JobEnum.BlackMage)
            elif job_name == "RedMage" : job_object = Player([], [DualCastEffect],{}, JobEnum.RedMage)
            elif job_name == "Summoner" : job_object = Player([], [],{}, JobEnum.Summoner)
            #Ranged
            elif job_name == "Dancer" : job_object = Player([], [EspritEffect],{}, JobEnum.Dancer)
            elif job_name == "Machinist" : job_object = Player([], [],{}, JobEnum.Machinist)
            elif job_name == "Bard" : job_object = Player([], [SongEffect],{}, JobEnum.Bard)
            #melee
            elif job_name == "Reaper" : job_object = Player([], [],{}, JobEnum.Reaper)
            elif job_name == "Monk" : job_object = Player([], [ComboEffect],{}, JobEnum.Monk)
            elif job_name == "Dragoon" : job_object = Player([], [],{}, JobEnum.Dragoon)
            elif job_name == "Ninja" : job_object = Player([], [],{}, JobEnum.Ninja)
            elif job_name == "Samurai" : job_object = Player([], [],{}, JobEnum.Samurai)
                    
                
            job_object.playerID = str(player["id"])
            player_list[str(player["id"])] = {"name" : player["name"], "job" : job_name, "job_object" : job_object} #Adding new Key
            #We can access the information using the player's id

    



    #2nd request will look at all prepull actions done by the players

    payload = "{\"query\":\"query trio{\\n\\treportData {\\n\\t\\treport(code: \\\""+fightID+"\\\") {\\n\\t\\t\\ttitle,\\n\\t\\t\\tendTime,\\n\\t\\t\\tevents(\\n\\t\\t\\t\\tendTime:1649370483952,\\n\\t\\t\\t\\tfightIDs:"+fightNumber+",\\n\\t\\t\\t\\tincludeResources: false,\\n\\t\\t\\t\\tfilterExpression:\\\"type = 'combatantinfo'\\\"\\n\\t\\t\\t){data\\n\\t\\t\\t}\\n\\t\\t\\t\\n\\t\\t}\\n\\t}\\n}\",\"operationName\":\"trio\"}"
    conn.request("POST", "/api/v2/client", payload, headers)
    res = conn.getresponse()
    data = res.read()
    data_json = json.loads(data.decode("utf-8"))

    combatantinfo_list = data_json["data"]["reportData"]["report"]["events"]["data"] #Array of combatantinfo

    for info in combatantinfo_list:
        sourceID = str(info["sourceID"])
        player_obj = player_list[sourceID]["job_object"]
        auras = info["auras"]

        for aura in auras: #Going through all buffs in the player
            #We will look for a selection of buffs that are important. We will assume
            #the optimal scenario. So if we use a potion, we will assume its right before the fight begins
            
            player_obj.auras += [aura["name"]] #Adding aura. Used for when restoring a fight using a saved file.
            
            if aura["name"] == "SharpCast":
                #SharpCast for BLM.
                player_obj.SharpCast = True
            elif aura["name"] == "Soulsow":
                player_obj.Soulsow = True
            elif aura["name"] == "Medicated":
                #Potion
                player_obj.EffectCDList.append(PrepullPotion) #Adding this effect that will automatically apply the potion
                #on the first go through of the sim
            elif aura["name"] == "Meikyo Shisui":
                player_obj.EffectCDList.append(MeikyoStackCheck)
                player_obj.MeikyoCD = 46
                player_obj.MeikyoStack -= 1
                player_obj.EffectList.append(MeikyoEffect)
                player_obj.EffectCDList.append(MeikyoCheck) #Could be a problem if do it before finishing 3 weaponskills
                player_obj.Meikyo = 3
            elif aura["name"] == "Mudra":
                #Will also assume huton has been done
                ApplyHuton(player_obj, None) #Giving Huton
                player_obj.HutonTimer = 600 #Assuming some loss
                #If mudra is detected, we will assume we have casted it for Suiton
                player_obj.CurrentRitual = [0,1,2]
            elif aura["name"] == "Eukrasia":
                player_obj.Eukrasia = True
            elif aura["name"] == "Standard Step":
                #Assuming Dancer did 2 step
                player_obj.Emboite = True
                player_obj.Entrechat = True
                player_obj.StandardFinish = True
            elif aura["name"] == "ClosedPosition":
                if dance_partner_flag:
                    player_obj.DancePartner = dance_partner
                else:
                    closed_position = True
                    dancer = player_obj
            elif aura["name"] == "Dance partner":
                if closed_position:
                    dancer.DancePartner = player_obj
                else:
                    dance_partner_flag = True
                    dance_partner = player_obj





    #Third request will fetch all the abilities done in the fight and make an array associated with each player's ID

    payload = "{\"query\":\"query trio{\\n    reportData {\\n        report(code: \\\""+fightID+"\\\") {\\n\\t\\t\\t\\tendTime,\\n            events(\\n\\t\\t\\t\\t\\t\\t\\tfightIDs:"+fightNumber+",\\n\\t\\t\\t\\t\\t\\t\\tendTime:99999999999999,\\n\\t\\t\\t\\t\\t\\t\\tincludeResources:false,\\n\\t\\t\\t\\t\\t\\t\\tfilterExpression:\\\"type = 'cast' OR type = 'begincast' OR type = 'calculateddamage' OR type = 'applybuff' or type = 'calculatedheal'\\\",\\n\\t\\t\\t\\t\\t\\t\\tlimit:10000\\n\\t\\t\\t\\t\\t\\t\\t\\n\\t\\t\\t\\t\\t\\t){data}\\n        }\\n\\n    }\\n}\",\"operationName\":\"trio\"}"
    conn.request("POST", "/api/v2/client", payload, headers)
    res = conn.getresponse()
    data = res.read()
    data_json = json.loads(data.decode("utf-8"))

    action_list = data_json["data"]["reportData"]["report"]["events"]["data"] #Array of all actions done


    class action_object():
        def __init__(self, action_id, timestamp, type, targetID, sourceID, extraAbilityGameID):
            self.action_id = action_id #actionID
            self.timestamp = timestamp #relative timestamp
            self.type = type
            self.targetID = targetID
            self.targetEnemy = targetID in enemy_list #if the targetID is one of the enemy ID, then we are targeting an enemy.
            self.targetSelf = targetID == sourceID #if the action targets the player casting it
            self.isHeal = type == "calculatedheal"
            self.extraAbilityGameID = extraAbilityGameID
            #If we are not, we will have to do some more stuff later on.

        def isequal(self, action): #This checks if two consecutive actions are equivalent. 
            return self.action_id == action.action_id and self.timestamp == action.timestamp and self.type == action.type

        def __str__(self):
            return "action_id : " + str(self.action_id) + " type : " + self.type + " timestamp : " + str(self.timestamp/1000)

    action_dict = {} #Each player can be found using their id, and each key has an array of all done actions in order. Each entry in the array is an action_object

    #Will first initialize all key in the action_dict to be all empty arrays with key equal to player's ID

    for key in player_list:
        action_dict[key] = [] #Initializing each array
    #print(action_dict.keys())
    for action in action_list:
        #Will parse each action so each player has a list of all done action
        if str(action["sourceID"]) in action_dict.keys(): #Making sure the sourceID is a player
            rel_timestamp = action["timestamp"] - relative_timestamp_zero
            if "extraAbilityGameID" in action.keys() and action["abilityGameID"] == 1000049: 
                action_dict[str(action["sourceID"])] += [action_object(action["abilityGameID"], rel_timestamp, action["type"], action["targetID"], action["sourceID"], action["extraAbilityGameID"])] #Special for tincture
            else : action_dict[str(action["sourceID"])] += [action_object(action["abilityGameID"], rel_timestamp, action["type"], action["targetID"], action["sourceID"], 0)]#Will be assumed to be damaging spell

    #Will now go through each player and build their action set
    for player in action_dict:
        player_action_list = [] #will contain the action list that we have to give to the sim for the player
        #player is each player's id
        wait_timestamp = 0 #timestamp value of an action so we can compare it to the next action to add WaitAbility
        raw_action_list = action_dict[str(player)] #list of action_object object
        wait_flag = False #a flag that is set to true if we have to add a WaitAbility() in the player_action_list
        wait_cast = False #a flag that is set to true if we are waiting for the next cast type
        wait_calculateddamage = False #a flag that is set to true if we are waiting for calculated damage
        is_casted = False #flag that is set to true if the spell is casted (the server has a shorter casting time cuz why not) so have to play around that
        is_heal = False #flag that is set to true if the spell is a healing spell
        previous_action = action_object(0, 0, "", 0, 0, 0)

        #Edge case flags
        in_barrage = False #Flag that is set to true if we are looking for the barrage pattern (for bard)
        once_through = False #A flag to know if we have gone through barrage once
        calculated_damage_couter = 0 #A counter to know how much calculated damage we have seen

        wait_potion = False #Flag to know if we wait one cycle for a tincture

        for action in raw_action_list:
            #will check the type since we have to do different stuff in accordance to what it is


            if not previous_action.isequal(action): #Checks if this new action is the same, if it is we skip it
                previous_action = action


                if wait_flag: #If the flag is True, we have to add a WaitAbility
                    wait_time = (action.timestamp - wait_timestamp)
                    if wait_time >=500: #otherwise animation lock
                        #player_action_list.append(WaitAbility(max(0,(wait_time))/ 1000)) #Dividing by 1000 since time in milisecond
                        wait_flag = False #reset
                        wait_timestamp = 0 #
                        
                
                if action.action_id == 1000049: #Tincture
                    action.action_id = action.extraAbilityGameID

                next_action = lookup_abilityID(action.action_id, action.targetID, player, player_list) #returns the action object of the specified spell
                
                if is_heal:
                    #If this flag is set to true, we wait until we do not have type = 'calculatedheal'
                    if type != 'calculatedheal':
                        is_heal = False


                if in_barrage:
                    #If barrage was the last ability, then we need to wait for 3 calculated damage in a row
                    #We will first check if the action is a weaponskill, which is required for barrage to work
                    #This is so if we do an oGCD after barrage, it still works
                    if once_through:
                        calculated_damage_couter += 1 #We do nothing
                        #input("here 2")
                    elif isinstance(next_action, BardSpell) and next_action.Weaponskill:
                        #If we get here, we have to wait for 1 cast and  3 calculated damage
                        #We will still let the action go through the selection, but for the next time we will intercept it here
                        once_through = True
                        calculated_damage_couter = 0
                        #input("here 1")
                    if calculated_damage_couter == 3: #If we have seen two other (for a total of 3)
                        #Then we are done with this edge case, so we reset all flags and counters
                        calculated_damage_couter = 0
                        once_through = False
                        in_barrage = False
                        #input("here 3")
                #We will check for edge cases so we do not have to worry about them
                elif action.action_id == 107: #Bard barrage
                    #It will follow with 1 cast and 3 calculated damage since it actually use 3 of the next weaponskill
                    #But the sim only needs to know which weaponskill is done after, so we will filter it
                    in_barrage = True #Setting flag to true]
                    #input("Detected barrage")
                elif action.action_id == 1000049: #Tincture
                    player_action_list.append(next_action)
                    wait_potion = True
                    wait_flag = True
                    wait_timestamp = action.timestamp



                if not is_heal and calculated_damage_couter < 1 and not wait_potion: #No edge cases. Just proceed normally
                    if not wait_cast and not wait_calculateddamage:
                        if action.type == "begincast":#If begining cast, we simply add the spell to the list
                            player_action_list.append(next_action)
                            wait_cast = True #set flag to true
                            is_casted = True #says the action is a casted action
                            #if isinstance(next_action, BardSpell) : input("Adding : " + str(next_action.id))
                        elif action.type == "cast":
                            if next_action != None: player_action_list.append(next_action)
                            wait_calculateddamage = True
                            wait_flag = True
                            wait_timestamp = action.timestamp
                            #if isinstance(next_action, BardSpell) : input("Adding : " + str(next_action.id))
                        elif action.type == "calculateddamage":#insta cast, so we want to add but also check how long until next action. Calculated damage might also be right after a "cast", so we want to have
                            #it such that it can detect if it is an "insta-cast" or the damage from a casted action (which will affect how we add it to the action_list)
                            player_action_list.append(next_action)
                            wait_flag = True #We have to add a WaitAbility, so we will check this time and next action's timestamp and add a relevant WaitAbility
                            wait_timestamp = action.timestamp
                            #if isinstance(next_action, BardSpell) : input("Adding : " + str(next_action.id))
                        elif action.type == "calculatedhealing":
                            #same as calculateddamage, but since its a healing, we want to make sure we do not add the action
                            #for each player it hits (if it is an AOE heal)
                            player_action_list.append(next_action)
                            wait_flag = True #We have to add a WaitAbility, so we will check this time and next action's timestamp and add a relevant WaitAbility
                            wait_timestamp = action.timestamp
                            is_heal = True
                    elif wait_cast:
                        #Waiting for a cast
                        if action.type == "cast":
                            wait_cast = False
                            wait_calculateddamage = True
                    elif wait_calculateddamage:
                        if action.type == "calculateddamage" or action.type == "applybuff":
                            wait_calculateddamage = False
                            wait_flag = True #Waiting on next action
                            if is_casted: 
                                wait_timestamp = 500 + action.timestamp
                                is_casted = False
                            else:wait_timestamp = action.timestamp
                        elif action.type == "calculateddamage": #Same as calculateddamage, but with heal check
                            wait_calculateddamage = False
                            wait_flag = True #Waiting on next action
                            if is_casted: 
                                wait_timestamp = 500 + action.timestamp
                                is_casted = False
                            else:wait_timestamp = action.timestamp
                            is_heal = True
                        elif action.type == "cast":
                            if next_action != None: 
                                player_action_list.append(next_action)

                wait_potion = False #We have waited for the potion

        action_dict[player] = player_action_list

    return action_dict, player_list
    #action_dict is a dictionnary with a list of actions done by each player. Accessible by using their IDs
    #player_list is a dictionnary with relevant information to all players. Accessible by using their IDs
#Function to test timing

def test(client_id,client_secret):
    conn = http.client.HTTPSConnection("www.fflogs.com")
    access_token = getAccessToken(conn, client_id, client_secret)
    actionID = 7

    payload = "{\"query\":\"query trio{\\n\\treportData {\\n\\t\\treport(code: \\\"RQwfx3vATFWGahJc\\\") {\\n\\t\\t\\ttitle,\\n\\t\\t\\tendTime,\\n\\t\\t\\tevents(\\n\\t\\t\\t\\tendTime:1651557651925,\\n\\t\\t\\t\\tfightIDs:8,\\n\\t\\t\\t\\tincludeResources: false,\\n\\t\\t\\t\\tfilterExpression:\\\"ability.ID = " + str(actionID) +"\\\"\\n\\t\\t\\t){data\\n\\t\\t\\t}\\n\\t\\t\\t\\n\\t\\t}\\n\\t}\\n}\",\"operationName\":\"trio\"}"
    
    headers = {
        'Content-Type': "application/json",
        'Authorization': "Bearer %s" % access_token
        }

    conn.request("POST", "/api/v2/client", payload, headers)

    res = conn.getresponse()
    data = res.read()
    data_json = json.loads(data.decode("utf-8"))

    data = data_json["data"]["reportData"]["report"]["events"]["data"]
    wait_cast = False
    wait_calculated = False
    for event in data:

        if event["type"] == "cast" and event["sourceID"] == 56: input("Applying fire 4 at : " + str((event["timestamp"] - 14825136)/1000))

        if event["type"] == "begincast": 
             begin_time = event["timestamp"]
             wait_cast = True
        elif event["type"] == "cast" and wait_cast:
             end_time = event["timestamp"]
             input("Fire 4 had a casttime of " + str((end_time - begin_time)/1000))
             wait_calculated = True
             wait_cast = False
        elif wait_calculated and event["type"] == "calculateddamage":
             wait_calculated = False




from ffxivcalc.Jobs.Base_Spell import DOTSpell, buff, empty, buffHistory, buffPercentHistory
from ffxivcalc.Jobs.Ranged.Ranged_Spell import BardSpell
from ffxivcalc.helperCode.helper_math import roundDown
import copy
Lock = 0

#Thanks for birdy for da help lol. Bitchiest bard I've ever known, much love

"""
Because of the inherent random nature of Bard, every spell that requires luck based proc
will be avaible at all time. However, the Player object will keep track of an expected amount of
procs and will compare it at the end to the number of actually done procs of that ability.
A check will be done notheless to make sure a rotation is at least possible given enough luck
"""


#Requirement

def RefulgentArrowRequirement(Player, Spell):
    return Player.StraightShotReady, -1

def SidewinderRequirement(Player, Spell):
    return Player.SidewinderCD <= 0, Player.SidewinderCD

def EmpyrealArrowRequirement(Player, Spell):
    return Player.EmpyrealArrowCD <= 0, Player.EmpyrealArrowCD
"""
def PitchPerfectRequirement(Player, Spell):
    #I will modify the potency of the spell here since if I do it in apply, it will be harder to access spell
    #I won't consider if only 1 repertoire since the spell has by default 100 potency
    if Player.MaximumRepertoire == 0 : return False
    elif Player.MaximumRepertoire == 2:
        Spell.Potency += 120
    elif Player.MaximumRepertoire == 3:
        Spell.Potency += 260
    else:
        input("lol, somehting went wrong in bard")
    return Player.WanderingMinuet
"""
def PitchPerfect1Requirement(Player, Spell):
    if Player.MaximumRepertoire < 1:#It will not be possible to cast in any case
        return False, -1
    else:
        #We can cast it, and we have to check the chances
        need = max(0, 1 - Player.ExpectedRepertoire)#This represents how much we would be missing, compared to the expected value
        Player.UsedRepertoireAdd += need#This represents the total number of repertoire procs we have used through out the simulation
        return Player.WanderingMinuet, -1

def PitchPerfect2Requirement(Player, Spell):
    if Player.MaximumRepertoire < 2:#It will not be possible to cast in any case
        return False, -1
    else:
        #We can cast it, and we have to check the chances
        need = max(0, 2 - Player.ExpectedRepertoire)#This represents how much we would be missing, compared to the expected value
        Player.UsedRepertoireAdd += need#This represents the total number of repertoire procs we have used through out the simulation
        return Player.WanderingMinuet, -1

def PitchPerfect3Requirement(Player, Spell):
    if Player.MaximumRepertoire < 3:#It will not be possible to cast in any case
        return True, -1
    else:
        #We can cast it, and we have to check the chances
        need = max(0, 3 - Player.ExpectedRepertoire) #This represents how much we would be missing, compared to the expected value
        Player.UsedRepertoireAdd += need #This represents the total number of repertoire procs we have used through out the simulation
        return Player.WanderingMinuet, -1

def WanderingMinuetRequirement(Player, Spell):
    return Player.WanderingMinuetCD <= 0, Player.WanderingMinuetCD

def BattleVoiceRequirement(Player, Spell):
    return Player.BattleVoiceCD <= 0, Player.BattleVoiceCD

def BloodLetterRequirement(Player, Spell):
    #This requirement will check if the spell can be casted. If it cannot, it will cast it anyway, but will add to
    #the UsedBloodLetterReduction so we know by how much we go over it
    if Player.BloodLetterStack <= 0:# In which case we have to add to UsedBloodLetterReduction
        #We will look at the current cooldown on the ability, and add the rest of what it needs
        #Might want to check and make it more punishing if we are not in mage ballad



        #We will first check if a blood letter at this point is even possible by looking at the maximal possible reduction
        need = 15 - Player.BloodLetterCD #What reduction we need
        #input("maxblood : " + str(Player.MaximumBloodLetterReduction))
        if Player.MaximumBloodLetterReduction - need < -10 : 
            pass
            #input("Bard rotation might not be feasible")
            #return False, -1 #If it is bigger, than this rotation is impossible


        Player.MaximumBloodLetterReduction -= need #Updating new MaximumBloodLetterReduction
        Player.UsedBloodLetterReduction += need

    return True, -1

def ArmyPaeonRequirement(Player, Spell):
    return Player.ArmyPaeonCD <= 0, Player.ArmyPaeonCD

def MageBalladRequirement(Player, Spell):
    return Player.MageBalladCD <= 0, Player.MageBalladCD

def BarrageRequirement(Player, Spell):
    return Player.BarrageCD <= 0, Player.BarrageCD

def RagingStrikeRequirement(Player, Spell):
    return Player.RagingStrikeCD <= 0, Player.RagingStrikeCD

def RadiantFinaleRequirement(Player, Spell):
    return Player.MageCoda or Player.ArmyCoda or Player.WandererCoda, -1

def BlastArrowRequirement(Player, Spell):
    return Player.BlastArrowReady, -1

def ShadowbiteRequirement(Player, Spell):
    return Player.ShadowbiteReady, -1

def TroubadourRequirement(Player, Spell):
    return Player.TroubadourCD <= 0, Player.TroubadourCD

def WardenPaeanRequirement(Player, Spell):
    return Player.WardenPaeanCD <= 0, Player.WardenPaeanCD

def NatureMinneRequirement(Player, Spell):
    return Player.NatureMinneCD <= 0, Player.NatureMinneCD

#Apply

def ApplyNatureMinne(Player, Enemy):
    Player.NatureMinneCD = 90

def ApplyWardenPaean(Player, Enemy):
    Player.WardenPaeanCD = 45

def ApplyTroubadour(Player, Enemy):
    Player.TroubadourCD = 90

def ApplyLadonsbite(Player, Enemy):
    Player.ShadowbiteReady = True
    Player.ExpectedShadowbite += 0.35

def ApplyShadowbite(Player, Enemy):
    Player.ShadowbiteReady = False
    Player.UsedShadowbite += 1

def ApplyBurstShot(Player, Enemy):
    Player.StraightShotReady = True #We assume it will be true, in reality it has 35% chance
    Player.ExpectedRefulgent += 0.35

def ApplyRefulgentArrow(Player, Enemy):
    Player.StraightShotReady = False
    Player.UsedRefulgent += 1

def ApplyStormbite(Player, Enemy):
    Player.StraightShotReady = True #We assume it will be true, in reality it has 35% chance
    Player.ExpectedRefulgent += 0.35
    if Player.StormbiteDOT == None:
        Player.StormbiteDOT = copy.deepcopy(StormbiteDOT)
        Player.DOTList.append(Player.StormbiteDOT)
        Player.EffectCDList.append(StormbiteDOTCheck)
    else : Player.StormbiteDOT.resetBuffSnapshot() # reset buff snapshot is reapplied
    Player.StormbiteDOTTimer = 45

def ApplyCausticbite(Player, Enemy):
    Player.StraightShotReady = True #We assume it will be true, in reality it has 35% chance
    Player.ExpectedRefulgent += 0.35
    if Player.CausticbiteDOT == None:
        Player.CausticbiteDOT = copy.deepcopy(CausticbiteDOT)
        Player.DOTList.append(Player.CausticbiteDOT)
        Player.EffectCDList.append(CausticbiteDOTCheck)
    else : Player.CausticbiteDOT.resetBuffSnapshot() # reset buff snapshot is reapplied
    Player.CausticbiteDOTTimer = 45

def ApplySidewinder(Player, Enemy):
    Player.SidewinderCD = 60

def ApplyIronJaws(Player, Enemy):
    Player.StraightShotReady = True #We assume it will be true, in reality it has 35% chance
    Player.ExpectedRefulgent += 0.35

    if Player.StormbiteDOT != None : Player.StormbiteDOTTimer = 45
    if Player.CausticbiteDOT != None : Player.CausticbiteDOTTimer = 45

def ApplyEmpyrealArrow(Player, Enemy):
    Player.EmpyrealArrowCD = 15

def ApplyPitchPerfect1(Player, Enemy):
    Player.MaximumRepertoire = max(0, Player.MaximumRepertoire - 1)
    Player.ExpectedRepertoire = max(0, Player.ExpectedRepertoire - 1)
    Player.UsedTotalWandererRepertoire += 1

def ApplyPitchPerfect2(Player, Enemy):
    Player.MaximumRepertoire = max(0, Player.MaximumRepertoire - 2)
    Player.ExpectedRepertoire = max(0, Player.ExpectedRepertoire - 2)
    Player.UsedTotalWandererRepertoire += 2

def ApplyPitchPerfect3(Player, Enemy):
    Player.MaximumRepertoire = max(0, Player.MaximumRepertoire - 3)
    Player.ExpectedRepertoire = max(0, Player.ExpectedRepertoire - 3)
    Player.UsedTotalWandererRepertoire += 3
    #input("Using PitchPerfect 3, max is : " + str(Player.MaximumRepertoire))


def ApplyBattleVoice(Player, Enemy):
    Player.BattleVoiceTimer = 15
    Player.BattleVoiceCD = 120
    Enemy.BattleVoice = True
    Player.EffectCDList.append(BattleVoiceCheck)

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffHistory(fight.TimeStamp, fight.TimeStamp + 15)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].BattleVoiceHistory.append(history)

def ApplyBloodLetter(Player, Enemy):
    if Player.BloodLetterStack == 3:
        Player.EffectCDList.append(BloodLetterStackCheck)
        Player.BloodLetterCD = 15
    Player.BloodLetterStack -= 1

def armyMuseCheck(Player, Enemy):
    if Player.ArmyMuseTimer <= 0:
        Player.hasteChangeValue = -Player.hasteGainedFromArmyMuse
        Player.Haste -= Player.hasteGainedFromArmyMuse 
        Player.hasteHasChanged = True
        Player.EffectToRemove.append(armyMuseCheck)

def _apply_army_muse(Player, Enemy):

    Player.ArmyMuseTimer = 10

    hasteBuff : int = 0
    repValue : int = int(Player.Repertoire)
    assert repValue >= 0 and repValue <= 4, "Player Repertoire value is lower than 0 or higher than 4"

    match repValue:
        case 0: hasteBuff = 0
        case 1 : hasteBuff = 1
        case 2 : hasteBuff = 2
        case 3 : hasteBuff = 4
        case 4 : hasteBuff = 12

    Player.hasteGainedFromArmyMuse = hasteBuff    
    Player.hasteChangeValue += hasteBuff
    Player.Haste = hasteBuff
    Player.hasteHasChanged = int(Player.hasteChangeValue) != 0 # Should never happend (not zero) but just in case.

    if _apply_army_muse in Player.EffectCDList : Player.EffectToRemove.append(_apply_army_muse)
    Player.EffectCDList.append(armyMuseCheck)

def checkArmyEthos(Player, Enemy):
    if Player.ArmyEthosTimer <= 0:
        Player.EffectToRemove.append(checkArmyEthos)
        
def _check_army_muse(Player):
    if Player.ArmyPaeon:
                             # Adding end of EffectCDList since ArmyPaeonCheck will remove its effect before it goes through this function.
        Player.EffectCDList.append(_apply_army_muse)
    elif checkArmyEthos in Player.EffectCDList:
        _apply_army_muse(Player, None)

        Player.ArmyEthosTimer = 0


def ApplyWanderingMinuet(Player, Enemy):
    Player.WandererCoda = True #Adding Coda
    Player.WanderingMinuetCD = 120
    Player.SongTimer = 45
    Enemy.WanderingMinuet = True
    Player.EffectCDList.append(WandererCheck)
    _check_army_muse(Player)
    #Removing Current song
    Player.MageBallad = False
    Player.ArmyPaeon = False
    Player.WanderingMinuet = True

def ApplyArmyPaeon(Player, Enemy):
    Player.Repertoire = 0 # Reseting ArmyPaeon repertoire count.
    Player.ArmyEthosTimer = 0 
    Player.hasteGainedFromArmyMuse = 0
    if _check_army_muse in Player.EffectCDList: Player.EffectCDList.remove(_check_army_muse)

    Player.ArmyCoda = True #Adding Coda
    Enemy.ArmyPaeon = True # 3% DH
    Player.ArmyPaeonCD = 120
    Player.SongTimer = 45
    Player.EffectCDList.append(ArmyPaeonCheck)
    Player.EffectList.append(ArmyPaeonEffect)
    #Have to remove current song
    Player.MageBallad = False
    Player.ArmyPaeon = True
    Player.WanderingMinuet = False

def ApplyMageBallad(Player, Enemy):
    Player.MageCoda = True #Adding Coda
    Enemy.buffList.append(MageBalladBuff)
    Player.SongTimer = 45
    Player.MageBalladCD = 120
    Player.EffectCDList.append(MageBalladCheck)
    _check_army_muse(Player)
    #Have to remove current song
    Player.MageBallad = True
    Player.ArmyPaeon = False
    Player.WanderingMinuet = False

def ApplyBarrage(Player, Enemy):
    Player.EffectList.append(BarrageEffect)
    Player.StraightShotReady = True #This one will always happen
    Player.ExpectedRefulgent += 1 #So we simply add 1 since ^
    Player.BarrageCD = 120

def ApplyRagingStrike(Player, Enemy):
    Player.RagingStrikeTimer = 20
    Player.buffList.append(RagingStrikeBuff)
    Player.EffectCDList.append(RagingStrikeCheck)

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20 , RagingStrikeBuff.MultDPS)
    #    Player.PercentBuffHistory.append(history)

def ApplyRadiantFinale(Player, Enemy):
    #Will first have to find how many coda we have
    coda = 0
    if Player.MageCoda: coda += 1
    if Player.ArmyCoda: coda += 1
    if Player.WandererCoda: coda += 1
    Player.RadiantFinaleBuff = copy.deepcopy(RadiantFinaleBuff)
    Player.RadiantFinaleBuff.MultDPS *= coda
    Player.RadiantFinaleBuff.MultDPS = round(1 + Player.RadiantFinaleBuff.MultDPS%1, 2)
    Player.RadiantFinaleTimer = 15
    Enemy.buffList.append(Player.RadiantFinaleBuff)
    Player.EffectCDList.append(RadiantFinaleCheck)
    Player.MageCoda = False
    Player.ArmyCoda = False
    Player.Wanderer = False
    #We used all coda

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 15 , Player.RadiantFinalBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplyBlastArrow(Player, Enemy):
    Player.BlastArrowReady = False

def ApplyApexArrow20(Player, Enemy):
    Player.UsedSoulVoiceGauge += 20

def ApplyApexArrow80(Player, Enemy):
    Player.UsedSoulVoiceGauge += 80
    Player.BlastArrowReady = True

#Effect

def BarrageEffect(Player, Spell):
    if not isinstance(Spell, DOTSpell) and Spell.Weaponskill:
        if Spell.id == Shadowbite.id: Spell.Potency += 100
        else : Spell.Potency *= 3 #Triples potency. Shouldn't be a problem since Bard has no combo
        Player.EffectToRemove.append(BarrageEffect)


def SongEffect(Player, Spell):
    Player.EffectCDList.append(SongEffectCheck)
    Player.EffectToRemove.append(SongEffect)
"""
def WandererEffect(Player, Spell):
    #This effect will keep track of how many repertoire we should be having
    input("increasing repertoire")
    if Player.SongTimer%3 == 0 and Player.SongTimer != 45:
        input("increasing repertoire")
        Player.MaximumRepertoire = min(3, Player.MaximumRepertoire + 1) #Adding to MaximumRepertoire, this is to make sure a Pitch Perfect is at least possible
        Player.ExpectedRepertoire = min(3, Player.ExpectedRepertoire + 0.8)
"""

def ArmyPaeonEffect(Player, Spell):
    if Spell.GCD: #This if is after since a spell can affect its own GCD
                             # There is no other source of haste for Bard. So to keep it simpler
                             # we will simply always set the value to whatever this is.
        pass
        #Spell.RecastTime *= (1 - 0.04 * Player.Repertoire) #Making GCD faster
        #Since CastTime is always Lock for Bard, only affecting RecastTime

#Check

def RadiantFinaleCheck(Player, Enemy):
    if Player.RadiantFinaleTimer <= 0:
        Enemy.buffList.remove(Player.RadiantFinaleBuff)
        Player.EffectToRemove.append(RadiantFinaleCheck)

def RagingStrikeCheck(Player, Enemy):
    if Player.RagingStrikeTimer <= 0:
        Player.buffList.remove(RagingStrikeBuff)
        Player.EffectToRemove.append(RagingStrikeCheck)

def StormbiteDOTCheck(Player, Enemy):
    if Player.StormbiteDOTTimer <= 0:
        Player.DOTList.remove(Player.StormbiteDOT)
        Player.StormbiteDOT = None
        Player.EffectToRemove.append(StormbiteDOTCheck)

def CausticbiteDOTCheck(Player, Enemy):
    if Player.CausticbiteDOTTimer <= 0:
        Player.DOTList.remove(Player.CausticbiteDOT)
        Player.CausticbiteDOT = None
        Player.EffectToRemove.append(CausticbiteDOTCheck)

def WandererCheck(Player, Enemy):


    #This effect is in the check since we want it to be called each frames
    if (int(Player.SongTimer*100)/100)%3 == 0 and Player.SongTimer != 45:
        Player.MaximumRepertoire = min(3, Player.MaximumRepertoire + 1) #Adding to MaximumRepertoire, this is to make sure a Pitch Perfect is at least possible
        Player.ExpectedRepertoire = min(3, Player.ExpectedRepertoire + 0.8)
        Player.ExpectedTotalWandererRepertoire += 0.8

    if Player.SongTimer <= 0 or not (Player.WanderingMinuet): #We check if timer, or if it becomes false because another song has begun
        Enemy.WanderingMinuet = False
        Player.WanderingMinuet = True
        Player.EffectToRemove.append(WandererCheck)

def ArmyPaeonCheck(Player, Enemy):

    #The effect will assume we get 4 procs in 5 GCD, since that is the expected number of proc it should take since we have 80% chance each GCD
    #We could change that as necessary
    #It will add 0.8 repertoire each proc since we have 80% chance for a max of 4 procs

    if Player.SongTimer != 45 and (int(Player.SongTimer*100)/100)%3 == 0 and not (Player.Repertoire == 4.0):
        Player.Repertoire += 0.8
        Player.hasteChangeValue = 3.2
        Player.Haste += 3.2
        Player.hasteHasChanged = True
        Player.hasteGainedFromArmyPaeon += 3.2

    if Player.SongTimer <= 0 or not (Player.ArmyPaeon):
        Enemy.ArmyPaeon = False
        Player.ArmyPaeon = False
        Player.EffectToRemove.append(ArmyPaeonCheck)
        Player.EffectList.remove(ArmyPaeonEffect)
        if _apply_army_muse not in Player.EffectCDList : Player.EffectCDList.append(checkArmyEthos)
        Player.ArmyEthosTimer = 30
        Player.hasteChangeValue = -Player.hasteGainedFromArmyPaeon
        Player.Haste -= Player.hasteGainedFromArmyPaeon 
        Player.hasteHasChanged = True

                             # Giving 

def MageBalladCheck(Player, Enemy):

    if Player.SongTimer != 45 and (int(Player.SongTimer*100)/100)%3 == 0: #The effect applies each 3 seconds, so we check each such interval 
        Player.ExpectedBloodLetterReduction += 7.5 * 0.8 #Adding expected reduction
        Player.MaximumBloodLetterReduction += 7.5

    if Player.SongTimer <= 0 or not (Player.MageBallad):
        Enemy.buffList.remove(MageBalladBuff) #Remove DPSBonus
        Player.MageBallad = False
        Player.EffectToRemove.append(MageBalladCheck)

def BattleVoiceCheck(Player, Enemy):
    if Player.BattleVoiceTimer <= 0:
        Player.EffectToRemove.append(BattleVoiceCheck)
        Enemy.BattleVoice = False

def BloodLetterStackCheck(Player, Enemy):
    if Player.BloodLetterCD <= 0:
        if Player.BloodLetterStack == 2:
            Player.EffectToRemove.append(BloodLetterStackCheck)
        else:
            Player.BloodLetterCD = 15
        Player.BloodLetterStack += 1


def SongEffectCheck(Player, Enemy):
    #This effect is constantly on the player and will keep track of the expected number of expected SoulVoiceGauge
    #This effect is in check since we want it to be called each frame
    if Player.SongTimer != 45 and (int(Player.SongTimer*100)/100)%3 == 0 and Player.SongTimer > 0 :
        Player.ExpectedSoulVoiceGauge += 5 * 0.8 #We have an expected of 5 voice each proc with a chance of 80%


#GCD
BurstShot = BardSpell(16495, True, 2.5, 220, ApplyBurstShot, [], True, type = 2)
RefulgentArrow = BardSpell(7409, True, 2.5, 280, ApplyRefulgentArrow, [RefulgentArrowRequirement], True, type = 2)
Stormbite = BardSpell(7407, True, 2.5, 100, ApplyStormbite, [], True, type = 2)
Causticbite = BardSpell(7406, True, 2.5, 150, ApplyCausticbite, [], True, type = 2)
StormbiteDOT = DOTSpell(-20, 25, True) #25
CausticbiteDOT = DOTSpell(-21, 20, True) #20
IronJaws = BardSpell(3560, True, 2.5, 100, ApplyIronJaws, [], True, type = 2)
ApexArrow20 = BardSpell(6, True, 2.5, 200, ApplyApexArrow20, [], True, type = 2)
ApexArrow80 = BardSpell(16496, True, 2.5, 500, ApplyApexArrow80, [], True, type = 2)
BlastArrow = BardSpell(25784, True, 2.5, 600, ApplyBlastArrow, [BlastArrowRequirement],True, type = 2)

#AOE GCD
Ladonsbite = BardSpell(25783, True, 0, 2.5, ApplyLadonsbite, [], True, type = 2)
Shadowbite = BardSpell(16494, True, 2.5, 170, ApplyShadowbite, [ShadowbiteRequirement], True, type = 2)


#Song
WanderingMinuet = BardSpell(3559, False, 0, 100, ApplyWanderingMinuet, [WanderingMinuetRequirement], False)
ArmyPaeon = BardSpell(116, False, 0, 100, ApplyArmyPaeon, [ArmyPaeonRequirement],False)
MageBallad = BardSpell(114, False, 0, 100, ApplyMageBallad, [MageBalladRequirement],False)
#oGCD
Sidewinder = BardSpell(3562, False, 0, 320, ApplySidewinder, [SidewinderRequirement],False)
EmpyrealArrow = BardSpell(3558, False, 0, 240, ApplyEmpyrealArrow, [EmpyrealArrowRequirement],False)
BattleVoice = BardSpell(118, False, 0, 0, ApplyBattleVoice, [BattleVoiceRequirement],False)
BloodLetter = BardSpell(110, False, 0, 110, ApplyBloodLetter, [BloodLetterRequirement],False)
RainOfDeath = BardSpell(117, False, 0, 110, ApplyBloodLetter, [BloodLetterRequirement],False) #AOE Version of BloodLetter
Barrage = BardSpell(107, False, 0, 0, ApplyBarrage, [BarrageRequirement],False)
RagingStrike = BardSpell(101, False, 0, 0, ApplyRagingStrike, [RagingStrikeRequirement],False)
RadiantFinale = BardSpell(25785, False, 0, 0, ApplyRadiantFinale, [RadiantFinaleRequirement],False)
Troubadour = BardSpell(7405, False, 0, 0, ApplyTroubadour, [TroubadourRequirement], False)
RepellingShot = BardSpell(112, False, 0, 0, empty, [], False) #No requirement, do that shit forever
WardenPaean = BardSpell(3561, False, 0, 0, ApplyWardenPaean, [WardenPaeanRequirement], False)
#Each PitchPerfecti represents a PitchPerfect with i repertoire
PitchPerfect1 = BardSpell(24, False, 0, 100, ApplyPitchPerfect1, [PitchPerfect1Requirement],False)
PitchPerfect2 = BardSpell(25, False, 0, 220, ApplyPitchPerfect2, [PitchPerfect2Requirement],False)
PitchPerfect3 = BardSpell(7404, False, 0, 360, ApplyPitchPerfect3, [PitchPerfect3Requirement],False)

#buff
RadiantFinaleBuff = buff(1.02,name="RF")
RagingStrikeBuff = buff(1.15,name="Raging Strike")
MageBalladBuff = buff(1.01,name="MB")

# Party mit buff
NatureMinne = BardSpell(7408, False, 0, 0, ApplyNatureMinne, [NatureMinneRequirement], False)

BardAbility = {
101 : RagingStrike, 
107 : Barrage, 
114 : MageBallad, 
116 : ArmyPaeon, 
117 : RainOfDeath, 
118 : BattleVoice, 
3558 : EmpyrealArrow, 
3559 : WanderingMinuet, 
3560 : IronJaws, 
3561 : WardenPaean, 
3562 : Sidewinder ,
7404 : PitchPerfect3, 
7405 : Troubadour, 
7406 : Causticbite, 
7407 : Stormbite, 
7408 : NatureMinne, 
7409 : RefulgentArrow, 
16494 : Shadowbite, 
16495 : BurstShot, 
16496 : ApexArrow80, 
25783 : Ladonsbite, 
25785 : RadiantFinale, 
112 : RepellingShot, 
110 : BloodLetter,
25784 : BlastArrow
}


"""
This file will contain the API module fastAPI. It is the file to launch if you
want to run the API.
"""

from fastapi import FastAPI
from ffxivcalc.API.Model.ObjectInModel import SimulateFightIn
from ffxivcalc.API.Model.ObjectOutModel import SimulateFightOut
from ffxivcalc.API.library import SimulateFightAPIHelper

# Make sure you have uvicorn installed and run the command
# python -m uvicorn API:app
# in the folder where API.py is located to run the API. Add --reload if you want the API to restart
# for every detected changes in the code.
# Then go to HOSTING_ADRESS/docs and you can tryout the different functionallities

app = FastAPI() # Creating api instance

@app.post("/SimulateFight", response_model=SimulateFightOut)
def GetSimulateFight(info : SimulateFightIn):
    """
    This API functionality lets someone ask for a simulation of a given fight.
    The API will request a JSON file with the correct format and will return the fight's
    results using as a schema SimulateFightOut.

    Args:\n
        info (SimulateFightIn): JSON file containing the fight's parameters.

    Returns:\n
        JSON : JSON file with the SimulateFightOut schema.
    """

    returnData = SimulateFightAPIHelper(info.dict())

    return returnData



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="ffxivcalc",  # Required
    version="0.8.940",  # Required
    description="DPS simulator for Final Fantasy XIV.",
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",  # Optional (see note above)
    url="https://github.com/IAmPythagoras/FFXIV-Combat-Simulator",  # Optiona
    author="Anthony Desrochers",
    author_email="anthony.desrochers17@gmail.com",
    classifiers=[  # Optional
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10"
    ],
    keywords="ffxiv, xiv, dps, simulator, ffxiv-combat-simulator",
    package_dir={"": "src"},  # Optional,
    packages=find_packages(where="src"),  # Required
    python_requires=">=3.10",
    install_requires=[  "contourpy",
                        "cycler",
                        "fonttools",
                        "kiwisolver",
                        "matplotlib",
                        "numpy",
                        "packaging",
                        "Pillow",
                        "pyparsing",
                        "python-dateutil",
                        "six",
                        "coreapi",
                        "pandas",
                        "python_graphql_client"],  # Optional
    project_urls={  # Optional
        "Source": "https://github.com/IAmPythagoras/FFXIV-Combat-Simulator",
        "Site": "https://ffxivdpscalc.azurewebsites.net/simulate/",
        "Documentation": "https://iampythagoras.github.io/index.html"
    },
)


from ffxivcalc.Jobs.Base_Spell import Potion, Spell, empty
from ffxivcalc.Jobs.Melee.Melee_Spell import ArmLength, SecondWind
Lock = 0
class RangedSpell(Spell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)


#########################################
########## MACHINIST SPELL  #############
#########################################


class MachinistSpell(RangedSpell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, WeaponSkill, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)

        self.Weaponskill = WeaponSkill #Boolean Variable


#########################################
#######    BARD    SPELL    #############
#########################################


class BardSpell(RangedSpell):

    def __init__(self, id, GCD, RecastTime, Potency, Effect, Requirement, Weaponskill, type = 0):
        super().__init__(id, GCD, Lock, RecastTime, Potency, 0, Effect, Requirement, type = type)

        self.Weaponskill = Weaponskill

#########################################
#########  DANCER    SPELL    ###########
#########################################


class DancerSpell(RangedSpell):

    def __init__(self, id, GCD, RecastTime, Potency, Effect, Requirement, Weaponskill, type = 0):
        super().__init__(id, GCD, Lock, RecastTime, Potency, 0, Effect, Requirement, type = type)
        self.Weaponskill = Weaponskill


#Class Action

#Requirement

def LegGrazeRequirement(Player, Spell):
    return Player.LegGrazeCD <= 0, Player.LegGrazeCD

def FootGrazeRequirement(Player, Spell):
    return Player.FootGrazeCD <= 0, Player.FootGrazeCD

def PelotonRequirement(Player, Spell):
    return Player.PelotonCD <= 0, Player.PelotonCD

def HeadGrazeRequirement(Player, Spell):
    return Player.HeadGrazeCD <= 0, Player.HeadGrazeCD

#Apply

def ApplyLegGraze(Player, Spell):
    Player.LegGrazeCD = 30

def ApplyFootGraze(Player, Spell):
    Player.FootGrazeCD = 30

def ApplyPeloton(Player, Spell):
    Player.PelotonCD = 5

def ApplyHeadGraze(Player, Spell):
    Player.HeadGrazeCD = 30


#ArmLength and SecondWind are in Melee_Spell.py
LegGraze = RangedSpell(0, False, Lock, 0, 0, 0, ApplyLegGraze, [LegGrazeRequirement])
FootGraze = RangedSpell(0, False, Lock, 0, 0, 0, ApplyFootGraze, [FootGrazeRequirement])
Peloton = RangedSpell(0, False, Lock, 0, 0, 0, ApplyPeloton, [PelotonRequirement])
HeadGraze = RangedSpell(0, False, Lock, 0, 0, 0, ApplyHeadGraze, [HeadGrazeRequirement])

# Limit Break actions

LB1Timer = 5.10
LB2Timer = 6.10
LB3Timer = 8.20

RangedLB1 = RangedSpell(1111, False,LB1Timer, LB1Timer, 0, 0, empty, [], type=3)
RangedLB2 = RangedSpell(1112, False,LB2Timer, LB2Timer, 0, 0, empty, [], type=3)
RangedLB3 = RangedSpell(1113, False,LB3Timer, LB3Timer, 0, 0, empty, [], type=3)

RangedAbility = {
7554 : LegGraze,
7553 : FootGraze,
7551: HeadGraze,
7557 : Peloton,
7541 : SecondWind,
7548 : ArmLength,
34590542 : Potion,
34592396 : Potion,
34594160 : Potion, # DEX pot ?
-2 : Potion,
1111 : RangedLB1,
1112 : RangedLB2,
1113 : RangedLB3
}


from ffxivcalc.Jobs.Base_Spell import buff, empty, DOTSpell
from ffxivcalc.Jobs.Tank.Tank_Spell import BigMitRequirement, ApplyBigMit, GunbreakerSpell
from ffxivcalc.Jobs.Player import MitBuff
import copy
Lock = 0.75


#Requirement

def CamouflageRequirement(Player, Spell):
    return Player.CamouflageCD <= 0, Player.CamouflageCD

def NoMercyRequirement(Player, Spell):
    return Player.NoMercyCD <= 0, Player.NoMercyCD

def RoughDivideRequirement(Player, Enemy):
    return Player.RoughDivideStack > 0, Player.RoughDivideCD

def BowShockRequirement(Player, Spell):
    return Player.BowShockCD <= 0, Player.BowShockCD

def SonicBreakRequirement(Player, Spell):
    return Player.SonicBreakCD <= 0, Player.SonicBreakCD

def DoubleDownRequirement(Player, Spell):
    return Player.DoubleDownCD <= 0, Player.DoubleDownCD

def HypervelocityRequirement(Player, Spell):
    return Player.ReadyToBlast, -1

def BloodfestRequirement(Player, Spell):
    return Player.BloodfestCD <= 0, Player.BloodfestCD

def BlastingZoneRequirement(Player, Spell):
    return Player.BlastingZoneCD <= 0, Player.BlastingZoneCD

def JugularRipRequirement(Player, Spell):
    return Player.ReadyToRip, -1

def AbdomenTearRequirement(Player, Spell):
    return Player.ReadyToTear, -1

def EyeGougeRequirement(Player, Spell):
    return Player.ReadyToGouge, -1

def AuroraRequirement(Player, Spell):
    return Player.AuroraStack > 0, Player.AuroraCD

def SuperbolideRequirement(Player, Spell):
    return Player.SuperbolideCD <= 0, Player.SuperbolideCD

def HeartOfLightRequirement(Player, Spell):
    return Player.HeartOfLightCD <= 0, Player.HeartOfLightCD

def HeartOfCorundumRequirement(Player, Spell):
    return Player.HeartOfCorundumCD <= 0, Player.HeartOfCorundumCD

#Apply

def ApplyCamouflage(Player, Enemy):
    Player.CamouflageCD = 90

    # Since its a 50% that additional 20% mit is added. We will simply 
    # assume a 20% total mit
    CamouflageMit = MitBuff(0.8, 20, Player)
    Player.MitBuffList.append(CamouflageMit)

def ApplyDemonSlice(Player, Enemy):
    if not (DemonSliceCombo in Player.EffectList) : Player.EffectList.append(DemonSliceCombo)

def ApplyHeartOfCorundum(Player, Enemy):
    Player.HeartOfCorundumCD = 25

def ApplyHeartOfLight(Player, Enemy):
    Player.HeartOfLightCD = 90

    # Will give every player 10% magic mit

    for player in Player.CurrentFight.PlayerList:
        player.MitBuffList.append(MitBuff(0.9, 15, player, MagicMit=True))

def ApplySuperbolide(Player, Enemy):
    Player.SuperbolideCD = 360

    Player.InvulnTimer = 10

def ApplyAurora(Player, Enemy):
    if Player.AuroraStack == 2:
        Player.EffectCDList.append(AuroraStackCheck)
        Player.AuroraCD = 60
    Player.AuroraStack -= 1

def ApplyNoMercy(Player, Enemy):
    Player.buffList.append(NoMercyBuff)
    Player.NoMercyTimer = 20
    Player.NoMercyCD = 60
    Player.EffectCDList.append(NoMercyCheck)

def ApplyRoughDivide(Player, Enemy):
    if Player.RoughDivideStack == 2:
        Player.EffectCDList.append(RoughDivideStackCheck)
        Player.RoughDivideCD = 30
    Player.RoughDivideStack -= 1

def ApplyBowShock(Player, Enemy):
    Player.BowShockCD = 60
    Player.BowShockTimer = 15
    if Player.BowShockDOT == None:
        Player.BowShockDOT = copy.deepcopy(BowShockDOT)
        Player.DOTList.append(Player.BowShockDOT)
        Player.EffectCDList.append(BowShockDOTCheck)
    else: Player.BowShockDOT.resetBuffSnapshot()

def ApplySonicBreak(Player, Enemy):
    Player.SonicBreakCD = 60
    Player.SonicBreakTimer = 30
    if Player.SonicBreakDOT == None:
        Player.SonicBreakDOT = copy.deepcopy(SonicBreakDOT)
        Player.DOTList.append(Player.SonicBreakDOT)
        Player.EffectCDList.append(SonicBreakDOTCheck)
    else: Player.SonicBreakDOT.resetBuffSnapshot()

def ApplyDoubleDown(Player, Enemy):
    Player.DoubleDownCD = 60 * Player.WeaponskillReduction

def ApplyHypervelocity(Player, Enemy):
    Player.ReadyToBurst = False

def ApplyBurstStrike(Player, Enemy):
    Player.ReadyToBlast = True

def ApplyBloodfest(Player, Enemy):
    Player.PowderGauge = 3
    Player.BloodfestCD = 90

def ApplyBlastingZone(Player, Enemy):
    Player.BlastingZoneCD = 30

def ApplyKeenEdge(Player, Enemy):
    if not (KeenEdgeCombo in Player.EffectList) : Player.EffectList.append(KeenEdgeCombo)

def ApplyGnashingFang(Player, Enemy):
    Player.ReadyToRip = True
    Player.GnashingFangCD = 30 * Player.WeaponskillReduction

def ApplySavageClaw(Player, Enemy):
    Player.ReadyToRip = False
    Player.ReadyToTear = True

def ApplyWickedTalon(Player, Enemy):
    Player.ReadyToTear = False
    Player.ReadyToGouge = True

def ApplyEyeGouge(Player, Enemy):
    Player.ReadyToGauge = False

#Combo Effect

def DemonSliceCombo(Player, Spell):
    if Spell.id == DemonSlaughter.id:
        Spell.Potency += 60
        Player.PowderGauge = min(3, Player.PowderGauge + 1)
        Player.EffectToRemove.append(DemonSliceCombo)

def KeenEdgeCombo(Player, Spell):
    if Spell.id == BrutalShell.id:
        Spell.Potency += 140
        if not (BrutalShellCombo in Player.EffectList) : Player.EffectList.append(BrutalShellCombo)
        Player.EffectToRemove.append(KeenEdgeCombo)

def BrutalShellCombo(Player, Spell):
    if Spell.id == SolidBarrel.id:
        Spell.Potency += 220
        Player.PowderGauge = min(3, Player.PowderGauge + 1)
        Player.EffectToRemove.append(BrutalShellCombo)


#Check

def NoMercyCheck(Player, Enemy):
    if Player.NoMercyTimer <= 0:
        Player.buffList.remove(NoMercyBuff)
        Player.EffectToRemove.append(NoMercyCheck)

def BowShockDOTCheck(Player, Enemy):
    if Player.BowShockTimer <= 0:
        Player.DOTList.remove(Player.BowShockDOT)
        Player.BowShockDOT = None
        Player.EffectToRemove.append(BowShockDOTCheck)

def SonicBreakDOTCheck(Player, Enemy):
    if Player.SonicBreakTimer <= 0:
        Player.DOTList.remove(Player.SonicBreakDOT)
        Player.SonicBreakDOT = None
        Player.EffectToRemove.append(SonicBreakDOTCheck)

def RoughDivideStackCheck(Player, Enemy):
    if Player.RoughDivideCD <= 0:
        if Player.RoughDivideStack == 1:
            Player.EffectToRemove.append(RoughDivideStackCheck)
        else:
            Player.RoughDivideCD = 30
        Player.RoughDivideStack +=1

def AuroraStackCheck(Player, Enemy):
    if Player.AuroraCD <= 0:
        if Player.AuroraStack == 1:
            Player.EffectToRemove.append(AuroraStackCheck)
        else:
            Player.AuroraCD = 30
        Player.AuroraStack +=1


#Combo Action

KeenEdge = GunbreakerSpell(16137, True, 2.5, 200, ApplyKeenEdge, [], 0, type = 2)
BrutalShell = GunbreakerSpell(16139, True, 2.5, 160, empty, [], 0, type = 2)
SolidBarrel = GunbreakerSpell(16145, True, 2.5, 140, empty, [], 0, type = 2)

GnashingFang = GunbreakerSpell(16146, True, 2.5, 380, ApplyGnashingFang, [], 1, type = 2)
JugularRip = GunbreakerSpell(16156, False, 0, 200, empty, [JugularRipRequirement], 0)
SavageClaw = GunbreakerSpell(16147, True, 2.5, 460, ApplySavageClaw, [JugularRipRequirement], 0, type = 2)
AbdomenTear = GunbreakerSpell(16157, False, 0, 240, empty, [AbdomenTearRequirement], 0)
WickedTalon = GunbreakerSpell(16150, True, 2.5, 540, ApplyWickedTalon, [AbdomenTearRequirement], 0, type = 2)
EyeGouge = GunbreakerSpell(16158, False, 0, 280, ApplyEyeGouge, [EyeGougeRequirement], 0)

BurstStrike = GunbreakerSpell(16162, True, 2.5, 380, ApplyBurstStrike, [], 1, type = 2)
Hypervelocity = GunbreakerSpell(25759, False, 0, 180, ApplyHypervelocity,[HypervelocityRequirement], 0)

#oGCD
BlastingZone = GunbreakerSpell(16165, False, 0, 720, ApplyBlastingZone, [BlastingZoneRequirement], 0)
Bloodfest = GunbreakerSpell(16164, False, 0, 0, ApplyBloodfest, [BloodfestRequirement], 0)
BowShock = GunbreakerSpell(16159, False, 0, 150, ApplyBowShock, [BowShockRequirement], 0)
BowShockDOT = DOTSpell(-10, 60, True)
RoughDivide = GunbreakerSpell(16154, False, 0, 150, ApplyRoughDivide, [RoughDivideRequirement], 0)
NoMercy = GunbreakerSpell(16138, False, 0, 0, ApplyNoMercy, [NoMercyRequirement], 0)
#GCD
DoubleDown = GunbreakerSpell(25760, True, 2.5, 1200, ApplyDoubleDown, [DoubleDownRequirement], 2, type = 2)
SonicBreak = GunbreakerSpell(16153, True, 2.5, 300, ApplySonicBreak, [SonicBreakRequirement], 0, type = 2)
SonicBreakDOT = DOTSpell(-9, 60,True)
LightningShot = GunbreakerSpell(16143, True, 2.5, 150, empty, [], 0, type = 2)
#AOE GCD
FatedCircle = GunbreakerSpell(16163, True, 2.5, 290, empty, [], 1, type = 2)
DemonSlice = GunbreakerSpell(16141, True, 2.5, 100, ApplyDemonSlice, [], 0, type = 2)
DemonSlaughter = GunbreakerSpell(16149, True, 2.5, 100, empty, [], 0, type = 2)


#Mit
Aurora = GunbreakerSpell(16151, False, 0, 0, ApplyAurora, [AuroraRequirement], 0)
Superbolide = GunbreakerSpell(16152, False, 0, 0, ApplySuperbolide, [SuperbolideRequirement], 0)
HeartOfLight = GunbreakerSpell(16160, False, 0, 0, ApplyHeartOfLight, [HeartOfLightRequirement], 0)
Camouflage = GunbreakerSpell(16140, False, 0, 0, ApplyCamouflage, [CamouflageRequirement], 0)
Nebula = GunbreakerSpell(16148, False, 0, 0, ApplyBigMit, [BigMitRequirement], 0)

def HeartOfCorundum(Target):
    """This function returns a GBNSpell object corresponding to HeartOfCorundum

    Args:
        Target (Player): Target of the action
    """

    def SpellApply(Player, Enemy):
        ApplyHeartOfCorundum(Player, Enemy)

        # Gives 15% mit for 8 seconds and 15% percent for 4 seconds
        # 8 seconds 15%
        CorundumMit = MitBuff(0.85, 8, Target)
        # Clarity of corundum 15% for 4s
        ClarityMit = MitBuff(0.85, 4, Target)

        Target.MitBuffList.append(CorundumMit)
        Target.MitBuffList.append(ClarityMit)

    HeartOfCorundum = GunbreakerSpell(25758, False, 0, 0, SpellApply, [HeartOfCorundumRequirement], 0)
    HeartOfCorundum.TargetID = Target.playerID
    return HeartOfCorundum






#buff
NoMercyBuff = buff(1.2,name="No Mercy")

GunbreakerAbility = {
16152 : Superbolide,
16143 : LightningShot,
16154 : RoughDivide,
16138 : NoMercy,
16164 : Bloodfest,
16137 : KeenEdge,
16139 : BrutalShell,
16145 : SolidBarrel,
16162 : BurstStrike,
16146 : GnashingFang,
16147 : SavageClaw,
16150 : WickedTalon,
16156 : JugularRip,
16157 : AbdomenTear,
16158 : EyeGouge,
25759 : Hypervelocity,
16153 : SonicBreak,
16165 : BlastingZone,
25760 : DoubleDown,
16141 : DemonSlice,
16149 : DemonSlaughter,
16163 : FatedCircle,
16148 : Nebula,
16140 : Camouflage,
25758 : HeartOfCorundum,
16151 : Aurora,
16160 : HeartOfLight,
16159 : BowShock
}


"""
This module contains all function used to verbally show a simulation's result to the user
and all other functions used to process the information offered by the library.   
"""
import math
import numpy as np
import logging
import matplotlib.pyplot as plt
from ffxivcalc.helperCode.Progress import ProgressBar
from ffxivcalc.helperCode.helper_math import roundDown
logging.getLogger('matplotlib').setLevel(logging.INFO) # silencing matplotlib logger
logging.getLogger('PIL').setLevel(logging.INFO) # silencing PIL logger


from ffxivcalc.Jobs.PlayerEnum import JobEnum

def SimulateRuns(fight, n : int, showBar : bool = True, pBNewBuffer = None):
    """
    This function will simulate the fight with ZIPActions the given number of time and will
    generate the DPS distribution from it
    n (int) -> Number of times to run the random simulation
    showBar : bool -> If true shows pB.
    pBNewBuffer : dict -> This will receive pB's currents frame in the key 'pb'
    """
    zipFightProgress = ProgressBar.init(n, "Computing DPS Dist")
    zipFightProgress.setShowBar(showBar)
    zipFightProgress.setExtraBuffer(pBNewBuffer)
    for i in range(n):
        fight.SimulateZIPFight()
        next(zipFightProgress)

    l = len(fight.PlayerList)
    fig, axs = plt.subplots((l // 4)+ (1 if l % 4 != 0 else 0), l if l < 4 else 4, constrained_layout=True) # DPS Crit distribution
    fig.suptitle("DPS Distribution (n = "+str(n)+" )")
    i = 0 # Used as coordinate for DPS distribution graph
    j = 0

    for player in fight.PlayerList:

        for runs in player.ZIPDPSRun:
            if str(runs) in player.DPSBar.keys():
                player.DPSBar[str(runs)] += 1
            else:
                player.DPSBar[str(runs)] = 1

        # ordering dict
        keys = list(player.DPSBar.keys())
        keys.sort()
        data = {i : player.DPSBar[i] for i in keys}

        Percent = int(n/100)
        percentile = [10,25,50,75,90,95,99]
        index = 0
        curTotal = 0
        for key in data:
            curTotal += data[key]
            if curTotal > (Percent * percentile[index]) : 
                player.ZIPRunPercentile[str(percentile[index])] = key
                index += 1
            if index == len(percentile) : break

        x = []
        y = []
        title = "" if player.PlayerName == "" else player.PlayerName + " ID - " + str(player.playerID)
        for bar in data:
            x += [float(bar)]
            y += [player.DPSBar[bar]/n]
        if l == 1:
            axs.plot(x, y)
            axs.plot([player.TotalDamage/fight.TimeStamp,player.TotalDamage/fight.TimeStamp], [0, 0.01])
            axs.set_ylim(ymin=0)
            axs.set_title(title)
        elif l <= 4:
            axs[i].plot(x, y)
            axs[i].plot([player.TotalDamage/fight.TimeStamp,player.TotalDamage/fight.TimeStamp], [0, 0.01])
            axs[i].set_ylim(ymin=0)
            axs[i].set_title(title)
        else:
            axs[j][i].plot(x, y)
            axs[j][i].plot([player.TotalDamage/fight.TimeStamp,player.TotalDamage/fight.TimeStamp], [0, 0.01])
            axs[j][i].set_ylim(ymin=0)
            axs[j][i].set_title(title)
        i+=1
        if i == 4:
            i = 0
            j+=1
    return fig
# Functions to print out all the results and plot DPS/PPS graph

def PrintResult(self, time : float, TimeStamp, PPSGraph : bool = True) -> str:
    """
    This function puts the result in a string that it will return which can then be printed
    self : Fight -> Fight we want the result of to be printed
    time : float -> final timestamp of the simulation
    TimeStamp : List[float] -> list of all timestamp where the DPS was saved in memory. Used to generate the graphs
    PPSGraph : bool -> If want the PPS graph next to the DPS graph
    """

    result_string = "The Fight finishes at: " + str(time) + "\n========================\n" 
    fig, axs = plt.subplots(1, 2 if PPSGraph else 1, constrained_layout=True) # DPS and PPS graph
    HPFig, HPaxs = plt.subplots(1,1, constrained_layout=True)

    HPaxs.set_ylabel("HP")
    HPaxs.set_xlabel("Time (s)")
    HPaxs.set_title("HP over time")
    HPaxs.spines["top"].set_alpha(0.0)
    HPaxs.spines["right"].set_alpha(0.0)
    HPaxs.set_facecolor("lightgrey")

    if PPSGraph:
        axs[0].set_ylabel("DPS")
        axs[0].set_xlabel("Time (s)")
        axs[0].set_title("DPS over time")
        axs[0].spines["top"].set_alpha(0.0)
        axs[0].spines["right"].set_alpha(0.0)
        axs[0].set_facecolor("lightgrey")
        axs[1].set_ylabel("PPS")
        axs[1].set_xlabel("Time (s)")
        axs[1].set_title("PPS over time")
        axs[1].spines["top"].set_alpha(0.0)
        axs[1].spines["right"].set_alpha(0.0)
        axs[1].set_facecolor("lightgrey")
    else:
        axs.set_ylabel("DPS")
        axs.set_xlabel("Time (s)")
        axs.set_title("DPS over time")
        axs.spines["top"].set_alpha(0.0)
        axs.spines["right"].set_alpha(0.0)
        axs.set_facecolor("lightgrey")

    fig.suptitle("Damage over time.")

    i = 0 # Used as coordinate for DPS distribution graph
    j = 0

    for player in self.PlayerList:
        HPaxs.plot(player.HPGraph[0], player.HPGraph[1], color="green")
        HPaxs.set_ylim(ymin=0)
        if time == 0: time = 1 # This is only so we don't have division by 0 error

        if player.TotalPotency == 0:
            PPS = 0
            DPS = 0
        else:
            PPS = player.TotalPotency / time
            DPS = player.TotalDamage / time
        
        result_string += (
            "Results for " + str(JobEnum.name_for_id(player.JobEnum)) + ("" if player.PlayerName == "" else " " + player.PlayerName) + " ID - " + str(player.playerID) + " :\n" + 
            "DPS : " + str(round(DPS,2)) + 
            " PPS : " + str(round(PPS,2)) + 
            " TP : " + str(round(player.TotalPotency,2)) + 
            " GCD : " + str(player.GCDCounter) + 
            " TD : " + str(player.TotalDamage) + "\n"
        )

        # DPS Percentile if applies

        result_string += ("DPS Percentile" + str(player.ZIPRunPercentile) + "\n") if len(player.ZIPRunPercentile.keys()) > 0 else ""


        # Plot part

        job = JobEnum.name_for_id(player.JobEnum)

        
        if player.JobEnum == JobEnum.Bard : 
            result_string += (
            "Procs/Gauge result (Used/Expecte) : \n" +
            "Refulgent : " + str(round(player.UsedRefulgent,2)) + "/" + str(round(player.ExpectedRefulgent,2)) + 
            " Wanderer Repertoire : " + str(round(player.UsedTotalWandererRepertoire,2)) + "/" + str(round(player.ExpectedTotalWandererRepertoire,2)) +
            " RepertoireAdd : " + str(round(player.UsedRepertoireAdd,2)) + 
            "\nSoul Voice : " + str(round(player.UsedSoulVoiceGauge,2)) + "/" + str(round(player.ExpectedSoulVoiceGauge,2)) +
            " Soul BloodLetter Reduction : " + str(round(player.UsedBloodLetterReduction,2)) + "/" + str(round(player.ExpectedBloodLetterReduction,2))
                             )
        elif player.JobEnum == JobEnum.Dancer:
            result_string += (
            "Procs/Gauge result (Used/Expected) : \n" +
            "Silken Symettry : " + str(round(player.UsedSilkenSymettry,2)) + "/" + str(round(player.ExpectedSilkenSymettry,2)) + 
            " Silken Flow : " + str(round(player.UsedSilkenFlow,2)) + "/" + str(round(player.ExpectedSilkenFlow,2)) +
            "\nFourfold Feather : " + str(round(player.UsedFourfoldFeather,2)) + "/" + str(round(player.ExpectedFourfoldFeather,2)) +
            " Threefold Fan : " + str(round(player.UsedThreefoldFan,2)) + "/" + str(round(player.ExpectedThreefoldFan,2))
                             )
        elif player.JobEnum == JobEnum.RedMage:
            result_string += (
            "Procs/Gauge result (Used/Expected) : \n" +
            "Verfire : " + str(round(player.UsedVerfireProc,2)) + "/" + str(round(player.ExpectedVerfireProc,2)) + 
            " Verstone : " + str(round(player.UsedVerstoneProc,2)) + "/" + str(round(player.ExpectedVerstoneProc,2)) 
                             )


        result_string += "\n=================\n"

        job_label = job + ("" if player.PlayerName == "" else (" " + player.PlayerName)) + " ID - " + str(player.playerID)
        if PPSGraph:
            axs[0].plot(TimeStamp,player.DPSGraph, label=job_label)
            axs[1].plot(TimeStamp,player.PotencyGraph, label=job_label)
        else:
            axs.plot(TimeStamp,player.DPSGraph, label=job_label)

        #if len(self.PlayerList) <= 8:
        #    if DPS != 0 : ComputeDPSDistribution(self, player, fig2, axs2[j][i], job)
        #    i+=1
        #    if i == 4:
        #        i = 0
        #        j+=1
    result_string += (
        "Total DPS : " + str(round(self.Enemy.TotalDamage / time, 2) if time != 0 else "0" ) + "\t" +
        "Total PPS : " + str(round(self.Enemy.TotalPotency/time,2) if time != 0 else "0" ) + "\t" +
        "Total Potency : " + str(round(self.Enemy.TotalPotency,2))
    )
    if PPSGraph: 
        axs[0].xaxis.grid(True)
        axs[1].xaxis.grid(True)
    else:
        axs.legend()
        
    if len(TimeStamp) == 0: # If for some reason Fight ended before 3 second TimeStamp is empty and we need to do an edge case to avoid a crash
        # Note that however the plot will be empty since no damage has been sampled
        if PPSGraph: 
            axs[0].xaxis.set_ticks(np.arange(0, 4, 25))
            axs[1].xaxis.set_ticks(np.arange(0, 4, 25))
        else:
            axs.xaxis.set_ticks(np.arange(0, 4, 25))
    else:
        if PPSGraph: 
            axs[0].xaxis.set_ticks(np.arange(0, max(TimeStamp)+1, 25))
            axs[1].xaxis.set_ticks(np.arange(0, max(TimeStamp)+1, 25))
        else:
            axs.xaxis.set_ticks(np.arange(0, max(TimeStamp)+1, 25))

    if PPSGraph: 
        axs[0].legend()
        axs[1].legend()
    else:
        axs.legend()

    return result_string, fig

